# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONWebKeySet`."""
from typing import Any, Iterable

import pydantic

from .ikeychain import IKeychain
from .jsonwebkey import JSONWebKey


class JSONWebKeySet(pydantic.BaseModel, IKeychain):
    __module__: str = 'ckms.types'
    keys: list[JSONWebKey] = []
    _index: dict[str, int] = pydantic.PrivateAttr()

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._index = {}
        for i, spec in enumerate(self.keys):
            if not spec.kid: # pragma: no cover
                continue
            self._index[spec.kid] = i

    def dict(self, **kwargs: Any) -> dict[str, Any]: # type: ignore pragma: no cover
        kwargs.setdefault('exclude_defaults', True)
        return super().dict(**kwargs)

    def filter( # type: ignore
        self,
        algorithm: str | None = None,
        kid: str | None = None,
        use: str | None = None,
        op: str | None = None
    ) -> 'JSONWebKeySet':
        return type(self)(keys=super().filter(algorithm, kid, use, op, self.keys))

    def get(self, kid: str | None) -> JSONWebKey:
        if not self.has(kid):
            raise KeyError(kid)
        assert kid is not None
        return self.keys[self._index[kid]]

    def get_encryption_algorithms(self) -> list[str]:
        return [x.alg for x in self.keys if (x.use == 'enc' and x.alg)]

    def get_encryption_keys(self) -> list[JSONWebKey]:
        return [x for x in self.keys if x.use == 'enc']

    def get_signing_algorithms(self) -> list[str]:
        return [x.alg for x in self.keys if (x.use == 'sig' and x.alg)]

    def get_signing_keys(self) -> list[JSONWebKey]:
        return [x for x in self.keys if x.use == 'sig']

    def json(self, **kwargs: Any) -> str: # type: ignore pragma: no cover
        kwargs.setdefault('exclude_defaults', True)
        return super().json(**kwargs)

    def has(self, kid: str | None) -> bool:
        return kid in self._index

    def __len__(self) -> int:
        return len(self.keys)

    def __iter__(self) -> Iterable[JSONWebKey]: # type: ignore
        return iter(self.keys)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import pytest

from ckms.core.models import KeySpecification
from ckms.types import CipherText
from .baseprovidertestcase import BaseProviderTestCase


class EncryptionTestCase(BaseProviderTestCase):
    __module__: str = 'ckms.core.types.provider.tests'

    @pytest.mark.asyncio
    async def test_encrypt_bytes(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        ct = await self.encrypt(spec, data)
        assert isinstance(ct, CipherText)

    @pytest.mark.asyncio
    async def test_decrypt_bytes(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        ct = await self.encrypt(spec, data)
        assert isinstance(ct, CipherText)
        assert await self.decrypt(spec, ct) == data

    @pytest.mark.asyncio
    async def test_decrypt_bytes_with_aad(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        if not spec.is_aead():
            pytest.skip()
        aad = b'foo'
        ct = await self.encrypt(spec, data, aad=aad)
        assert isinstance(ct, CipherText)
        assert ct.aad == aad, ct.aad
        assert await self.decrypt(spec, ct) == data

    @pytest.mark.asyncio
    async def test_decrypt_bytes_with_missing_aad_raises_exception(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        if not spec.is_aead():
            pytest.skip()
        aad = b'foo'
        ct = await self.encrypt(spec, data, aad=aad)
        assert isinstance(ct, CipherText)
        assert ct.aad == aad, ct.aad
        ct.aad = None
        try:
            assert await self.decrypt(spec, ct) == data
        except self.provider.Undecryptable:
            pass

    @pytest.mark.asyncio
    async def test_decrypt_bytes_with_different_aad_raises_exception(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        if not spec.is_aead():
            pytest.skip()
        aad = b'foo'
        ct = await self.encrypt(spec, data, aad=aad)
        assert isinstance(ct, CipherText)
        assert ct.aad == aad, ct.aad
        ct.aad = b'bar'
        try:
            assert await self.decrypt(spec, ct) == data
        except self.provider.Undecryptable:
            pass

    @pytest.mark.asyncio
    async def test_decrypt_bytes_with_missing_iv_raises_exception(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        if not spec.is_symmetric():
            pytest.skip()
        ct = await self.encrypt(spec, data)
        assert isinstance(ct, CipherText)
        ct.iv = None
        try:
            await self.decrypt(spec, ct)
        except self.provider.Undecryptable:
            pass

    @pytest.mark.asyncio
    async def test_decrypt_bytes_with_different_iv_raises_exception(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        if not spec.is_asymmetric():
            pytest.skip()
        ct = await self.encrypt(spec, data)
        assert isinstance(ct, CipherText)
        ct.iv = os.urandom(12)
        try:
            assert await self.decrypt(spec, ct) == data
        except self.provider.Undecryptable:
            pass


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Any
from typing import Literal
from typing import Sequence

import pydantic
from pyasn1.type import univ
from pyasn1_modules import rfc5280
from typing_extensions import Annotated

from ckms.ext import asn1
from .basecertificateextension import BaseCertificateExtension
from .basicconstraints import BasicConstraints
from .extendedkeyusagetype import ExtendedKeyUsageType
from .generalname import Name
from .generalname import GeneralName


class SubjectAltNames(BaseCertificateExtension):
    ext_id: Literal['2.5.29.17'] = pydantic.Field('2.5.29.17', alias='extnID')
    names: list[Name] = []

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ):
        obj = values.pop('extnValue', None)
        names: list[Any] = values.setdefault('names', [])
        if obj is not None:
            value = cast(
                list[rfc5280.GeneralNames],
                asn1.decode_ber(obj, rfc5280.SubjectAltName())
            )
            for altname in value:
                name = cast(str, altname.getName()) # type: ignore
                value = altname.getComponentByName(name) # type: ignore
                names.append(GeneralName.parse_obj({'kind': name, 'value': value}))
        return values

    def get_encoded_extension(self) -> univ.OctetString:
        names = rfc5280.GeneralNames()
        names.extend([x.asn1() for x in self.names]) # type: ignore
        return asn1.encode_ber(names, rfc5280.SubjectAltName())


class KeyUsage(BaseCertificateExtension):
    ext_id: Literal['2.5.29.15'] = pydantic.Field(default='2.5.29.15', alias='extnID')
    mask: rfc5280.KeyUsage = rfc5280.KeyUsage.fromOctetString(b'\x00') # type: ignore

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        if values.get('extnValue'):
            values =  {
                **values,
                'mask': asn1.decode_der(
                    value=values.pop('extnValue'),
                    spec=rfc5280.KeyUsage()
                )
            }
        return values

    def get(self, index: int) -> bool:
        """Get the specified bit on the :class:`~pyasn1_modules.rfc5280.KeyUsage()`
        instance.
        """
        mask: Sequence[int] = cast(Sequence[int], tuple(self.mask))
        return (len(mask) + 1) >= index and bool(mask[index])

    def get_encoded_extension(self) -> univ.OctetString:
        return asn1.encode_ber(self.mask, rfc5280.KeyUsage())

    def set(self, index: int, value: int) -> None:
        """Set the specified bit on the :class:`~pyasn1_modules.rfc5280.KeyUsage()`
        instance.
        """
        mask: list[int] = list(tuple(self.mask)) # type: ignore
        mask[index] = value
        self.mask = rfc5280.KeyUsage(tuple(mask))

    class Config:
        arbitrary_types_allowed: bool = True


class ExtendedKeyUsage(BaseCertificateExtension):
    ext_id: Literal['2.5.29.37'] = pydantic.Field('2.5.29.37', alias='extnID')
    uses: list[ExtendedKeyUsageType] = pydantic.Field(..., alias='extnValue')

    @pydantic.validator('uses', pre=True)
    def preprocess_uses(
        cls,
        value: list[str] | univ.OctetString
    ) -> list[str]:
        if isinstance(value, univ.OctetString):
            content = asn1.decode_der(
                value=value,
                spec=rfc5280.ExtKeyUsageSyntax()
            )
            value = [str(x) for x in content] # type: ignore
        return value

    def get_encoded_extension(self) -> univ.OctetString:
        obj = rfc5280.ExtKeyUsageSyntax()
        obj.extend([rfc5280.KeyPurposeId(x.value) for x in self.uses]) # type: ignore
        return asn1.encode_ber(obj)


class CertificateExtension(pydantic.BaseModel):
    __root__: Annotated[
        BasicConstraints |
        SubjectAltNames |
        KeyUsage |
        ExtendedKeyUsage,
        pydantic.Field(discriminator='ext_id')
    ]

    @classmethod
    def parse_obj(cls, obj: Any) -> BaseCertificateExtension:
        return super().parse_obj(obj).__root__


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from typing import Any
from typing import Literal

import pydantic
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from .localkeyspecification import LocalKeySpecification


class LocalKey(pydantic.BaseModel):
    path: str | pathlib.Path
    password: bytes | None
    encoding: Literal['pem'] = 'pem'

    async def setup(self, spec: LocalKeySpecification) -> Any:
        with open(self.path, 'rb') as f:
            return load_pem_private_key(f.read(), self.password)


from typing import Any
from typing import Literal

import pydantic
from pyasn1.type import univ
from pyasn1_modules import rfc5280

from ckms.ext import asn1
from .basecertificateextension import BaseCertificateExtension


class BasicConstraints(BaseCertificateExtension):
    ext_id: Literal['2.5.29.19'] = pydantic.Field('2.5.29.19', alias='extnID')
    is_ca: bool = pydantic.Field(False, alias='cA')
    path_length: int | None = pydantic.Field(None, alias='pathLenConstraint')

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        if 'extnValue' in values:
            content = asn1.decode_der(
                value=values.pop('extnValue'),
                spec=rfc5280.BasicConstraints()
            )
            values.update(dict(content)) # type: ignore
        return values

    @pydantic.validator('path_length', pre=True)
    def validate_path_length(
        cls,
        value: int | univ.Integer | None
    ) -> int | None:
        if isinstance(value, univ.Integer):
            value = int(value) if value.isValue else None
        return value

    def get_encoded_extension(self) -> univ.OctetString:
        obj = rfc5280.BasicConstraints()
        obj['cA'] = univ.Boolean(self.is_ca)
        if self.path_length is not None:
            obj['pathLenConstraint'] = univ.Integer(self.path_length)
        return asn1.encode_ber(obj)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONB64`"""
from typing import Callable
from typing import Generator

from ckms.utils import b64decode
from ckms.utils import b64encode


class OctetB64(bytes):
    """A :mod:`pydantic` field type that serializes from URL-safe, base64-encoded
    input data.
    """
    __module__: str = 'ckms.core.models'

    @classmethod
    def __get_validators__(
        cls
    ) -> Generator[Callable[..., 'OctetB64'], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: bytes) -> 'OctetB64':
        if isinstance(value, cls):
            return value

        return cls(b64decode(value))

    def __repr__(self) -> str: # pragma: no cover
        return f'OctetB64({super().__repr__()})'

    def serialize(self) -> str:
        return bytes.decode(b64encode(self), 'ascii')


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import secrets
from typing import Any
from typing import Generator

from cryptography.hazmat.primitives.ciphers.algorithms import AES

from .decrypter import Decrypter
from .encrypter import Encrypter
from .ikeyspecification import IKeySpecification
from .iprovider import IProvider


class IContentEncryptionKey(Decrypter, Encrypter):
    __module__: str = 'ckms.types'
    algorithms: dict[str, dict[str, int]] = {
        'A128GCM': {'nbytes': 16},
        'A192GCM': {'nbytes': 24},
        'A256GCM': {'nbytes': 32},
        'A128CBC': {'nbytes': 16},
        'A192CBC': {'nbytes': 24},
        'A256CBC': {'nbytes': 32},
    }
    __spec: IKeySpecification

    @property
    def algorithm(self) -> str:
        return self.__spec.algorithm

    @property
    def iv(self) -> bytes:
        return self.__iv

    @property
    def provider(self) -> IProvider:
        return self.__spec.provider

    @property
    def spec(self) -> IKeySpecification:
        return self.__spec

    @classmethod
    def generate(cls, algorithm: str) -> 'IContentEncryptionKey':
        """Generate a new :class:`ContentEncryptionKey`."""
        raise NotImplementedError

    def __init__(self, spec: IKeySpecification):
        self.__spec = spec
        self.__iv = secrets.token_bytes(12)

    def get_private_key(self) -> AES:
        return self.__spec.get_private_key()

    def is_aead(self) -> bool:
        raise NotImplementedError

    def __await__(self) -> Generator[Any, None, 'IContentEncryptionKey']:
        async def f():
            if not self.__spec.is_loaded():
                await self.__spec
            return self
        return f().__await__()

    def __bool__(self) -> bool:
        return bool(self.__spec)

    def __bytes__(self) -> bytes:
        return self.__spec.get_private_bytes()

    def __len__(self) -> int:
        return self.algorithms[self.spec.algorithm]['nbytes'] # type: ignore


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`RemoteBlobParams`."""
from typing import Any
from typing import Generator

import pydantic


class RemoteBlobParams(pydantic.BaseModel):
    __module__: str = 'ckms.types'
    provider: str
    params: dict[str, Any] = {}

    def __init__(self, **kwargs: Any):
        params = kwargs.setdefault('params', {})
        for name in list(kwargs.keys()):
            if name in self.__fields__:
                continue
            params[name] = kwargs.pop(name)
        super().__init__(**kwargs)

    def dict(self, *args: Any, **kwargs: Any) -> dict[str, Any]: # type: ignore
        data = super().dict(*args, **kwargs)
        data.update(data.pop('params', {}))
        return data

    async def get_content(self) -> bytes:
        """Return a byte-sequence returning the content of the
        remote blob.
        """
        raise NotImplementedError

    def __await__(self) -> Generator[Any, None, bytes]: # pragma: no cover
        return self.get_content().__await__()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from unimatrix.exceptions import CanonicalException


__all__: list[str] = [
    'ClaimTypeError',
    'Malformed',
    'MalformedHeader',
    'MalformedObject',
    'MalformedPayload',
    'MissingProtectedClaim',
    'MissingProtectedHeader',
    'MissingRequiredClaim',
]


class Malformed(CanonicalException):
    __module__: str = 'ckms.core.types'
    http_status_code: int = 400


class MalformedObject(Malformed):
    code: str = 'MALFORMED_JOSE_OBJECT'
    message: str = (
        "The application refuses to complete the request because the JOSE "
        "object was malformed and could not be parsed."
    )


class MalformedHeader(Malformed):
    code: str = 'MALFORMED_JOSE_HEADER'
    message: str = (
        "The application refuses to complete the request because the JOSE "
        "header of the attached object was malformed."
    )


class MalformedPayload(Malformed):
    code: str = "MALFORMED_JOSE_PAYLOAD"
    message: str = (
        "The payload of the JOSE object included in the request is "
        "malformed and could not be interpreted by the application."
    )


class ClaimTypeError(Malformed):
    code: str = 'CLAIM_TYPE_ERROR'
    message: str = (
        "The data type of claim violated the specification."
    )

    def __init__(self, claim: str, **kwargs: Any):
        super().__init__(
            detail=(
                f"The '{claim}' claim contains an invalid data type."
            ),
            **kwargs
        )


class MissingProtectedHeader(Malformed):
    http_status_code: int = 403
    code: str = 'MISSING_PROTECTED_HEADER'
    message: str = (
        "The JSON Web Signature (JWS) did not include a protected header "
        "for one or more signatures."
    )


class MissingProtectedClaim(Malformed):
    claim: str

    def __init__(self, name: str):
        self.claim = name
        super().__init__( # type: ignore
            message="A required claim was missing from the protected JOSE header.",
            detail=(
                f"The '{name}' claim was not included in the JOSE header and "
                "the application requires it to succcesfully validate the token."
            )
        )


class MissingRequiredClaim(Malformed):
    code: str = "CLAIM_REQUIRED"
    claim: str

    def __init__(self, name: str):
        self.claim = name
        super().__init__( # type: ignore
            message="A required claim was missing from the JSON Web Token (JWT).",
            detail=(
                f"The '{name}' claim was not included in the JWT and "
                "the application requires it to succcesfully validate the token."
            )
        )




# pylint: skip-file
import base64
import binascii
import datetime
import json
import hashlib
from typing import Any

from cryptography.hazmat.primitives.asymmetric import utils


def b64encode_json(obj: dict[str, Any]) -> bytes:
    return b64encode(json.dumps(obj, sort_keys=True), 'utf-8')


def b64encode(buf: bytes | str, encoding: str = 'utf-8') -> bytes: # pragma: no cover
    if isinstance(buf, str):
        buf = str.encode(buf, encoding=encoding)
    return base64.urlsafe_b64encode(buf).replace(b"=", b"")


def b64encode_sha1(value: bytes) -> bytes:
    return b64encode(hashlib.sha1(value).digest())


def b64encode_sha256(value: bytes) -> bytes:
    return b64encode(hashlib.sha256(value).digest())


def b64encode_str(*args: Any, **kwargs: Any) -> str:
    return bytes.decode(b64encode(*args, **kwargs), 'ascii')


def b64encode_int(value: int) -> bytes: # pragma: no cover
    return b64encode(int.to_bytes(value, (value.bit_length() + 7) // 8, 'big'))


def b64decode(buf: bytes | str) -> bytes: # pragma: no cover
    if isinstance(buf, str):
        buf = buf.encode("ascii")
    rem = len(buf) % 4
    if rem > 0:
        buf += b"=" * (4 - rem)
    return base64.urlsafe_b64decode(buf)


def b64decode_json(
    buf: bytes | str,
    encoding: str = 'utf-8',
    on_failure: BaseException | None = None,
    require: type[list[Any]] | type[dict[str, Any]] | None = None
) -> dict[str, Any] | list[Any]:
    """Deserialize a Base64-encoded string or byte-sequence as JSON."""
    try:
        result = json.loads(bytes.decode(b64decode(buf), encoding))
        if not isinstance(result, (require,) if require else (dict, list)):
            raise ValueError 
    except ValueError as exception:
        raise (on_failure or exception)
    return result


def b64decode_int(value: bytes | str) -> int:
    return bytes_to_number(b64decode(value))


def current_timestamp() -> int:
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())


def number_to_bytes(value: int, l: int) -> bytes: # pragma: no cover
    padded = str.encode("%0*x" % (2 * l, value), "ascii")
    return binascii.a2b_hex(padded)


def bytes_to_number(value: bytes) -> int: # pragma: no cover
    return int(binascii.b2a_hex(value), 16)


def normalize_ec_signature(l: int, sig: bytes): # pragma: no cover
    r, s = utils.decode_dss_signature(sig)
    return number_to_bytes(r, l) + number_to_bytes(s, l)


def certificate_chain(pem: bytes) -> list[bytes]:
    """Split a PEM-encoded certificate chain into a list of
    individual certificates.
    """
    linesep = b'\n'
    start_line = b'-----BEGIN CERTIFICATE-----' + linesep
    cleaned_data = bytes.join(linesep, [
        x for x in bytes.split(pem, linesep)
        if not x.startswith(b'#')
    ])
    return [
        (start_line + x)
        for x in bytes.split(cleaned_data, start_line) if x
    ]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TypeAlias
from typing import Union

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519


CryptographyPublicKeyType: TypeAlias = Union[
    ec.EllipticCurvePublicKey,
    ed448.Ed448PublicKey,
    ed25519.Ed25519PublicKey,
    rsa.RSAPublicKey,
    x448.X448PublicKey,
    x25519.X25519PublicKey
]



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any
from typing import Sequence

from pyasn1.type import tag
from pyasn1.type import univ
from pyasn1_modules import rfc2986

from .attribute import Attribute


class AttributeList(list[Attribute]):
    _index: dict[str, int]

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: Sequence[Attribute] | Sequence[dict[str, Any]] | univ.SetOf
    ) -> 'AttributeList':
        if isinstance(value, rfc2986.Attributes):
            value = [
                Attribute.parse_obj(x)
                for x in value # type: ignore
            ]
        for i, item in enumerate(value):
            if isinstance(item, Attribute):
                continue
            value[i] = Attribute.parse_obj(item) # type: ignore
        return cls(value or []) # type: ignore

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._index = {x.type: i for i, x in enumerate(self)}

    def add(self, obj: Attribute) -> None:
        if str(obj.type) in self._index:
            i = self._index[obj.type]
            f = functools.partial(self.__setitem__, i)
        else:
            f = self.append
            i = len(self)
            self._index[obj.type] = i
        return f(obj)

    def asn1(self) -> univ.SetOf:
        obj = rfc2986.Attributes().subtype( # type: ignore
            implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
        )
        obj.extend([x.asn1() for x in self]) # type: ignore
        return obj

    def get(self, oid: str) -> Attribute:
        return self[self._index[oid]] # type: ignore

    def has(self, oid: str) -> bool:
        return oid in self._index

    def pop(self, oid: str) -> Attribute: # type: ignore
        return super().pop(self._index[oid])



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Operation`."""
from typing import Any
from typing import AsyncGenerator

from .digest import Digest
from .ikeyspecification import IKeySpecification
from .message import Message
from .plaintext import PlainText


class Operation:
    """The base class for all cryptographic operations."""
    __module__: str = 'ckms.core.types'
    __abstract__: bool = True
    content: Digest | Message | PlainText
    spec: Any
    presets: dict[str, Any] = {
        'A128KW'        : {'mode': 'KEK'},
        'A192KW'        : {'mode': 'KEK'},
        'A256KW'        : {'mode': 'KEK'},
        'A128GCM'       : {'mode': 'GCM'},
        'A192GCM'       : {'mode': 'GCM'},
        'A256GCM'       : {'mode': 'GCM'},
        'A128GCMKW'     : {'mode': 'GCM'},
        'A192GCMKW'     : {'mode': 'GCM'},
        'A256GCMKW'     : {'mode': 'GCM'},
        'EdDSA'         : {},
        'ES256'         : {'digest': 'sha256'},
        'ES384'         : {'digest': 'sha384'},
        'ES512'         : {'digest': 'sha512'},
        'ES256K'        : {'digest': 'sha256'},
        'HS256'         : {},
        'HS384'         : {},
        'HS512'         : {},
        'PS256'         : {'digest': 'sha256', 'padding': 'PSS'},
        'PS384'         : {'digest': 'sha384', 'padding': 'PSS'},
        'PS512'         : {'digest': 'sha512', 'padding': 'PSS'},
        'RS256'         : {'digest': 'sha256', 'padding': 'PKCS1v15'},
        'RS384'         : {'digest': 'sha384', 'padding': 'PKCS1v15'},
        'RS512'         : {'digest': 'sha512', 'padding': 'PKCS1v15'},
        'RSA1_5'        : {'padding': 'PKCS1v15', 'digest': 'sha1'},
        'RSA-OAEP'      : {'padding': 'OAEP', 'digest': 'sha1'},
        'RSA-OAEP-256'  : {'padding': 'OAEP', 'digest': 'sha256'},
        'RSA-OAEP-384'  : {'padding': 'OAEP', 'digest': 'sha384'},
        'RSA-OAEP-512'  : {'padding': 'OAEP', 'digest': 'sha512'},
    }

    @property
    def digestmod(self) -> str | None:
        return self.content.digestmod

    @staticmethod
    def can_use_key(key: Any) -> bool:
        """Return a boolean indicating if the given `key` can be used to
        perform the operation.
        """
        return True

    @classmethod
    def fromalgorithm(
        cls,
        *,
        algorithm: str,
        **kwargs: Any
    ) -> 'Operation':
        """Create a new :class:`Operation` from a named algorithm."""
        if cls.presets:
            if algorithm not in cls.presets:
                raise NotImplementedError(f'No parameters known for {algorithm}.')
            kwargs.update(cls.presets[algorithm])
        return cls(**kwargs)

    def __init__(
        self,
        *,
        spec: Any,
        content: Digest | Message | PlainText,
        **kwargs: Any
    ):
        self.spec = spec
        if not self.can_use_key(self.spec): # pragma: no cover
            raise ValueError("Unsuitable key.")
        self.content = content

    def get_keyspec(self) -> IKeySpecification:
        return self.spec

    async def get_message(self) -> bytes:
        """Return the message on which the operation is performed."""
        message = b''
        async for chunk in self.chunks():
            message += chunk
        return message

    async def get_digest(self) -> bytes:
        """Return a digest of the message."""
        raise NotImplementedError

    def chunks(self) -> AsyncGenerator[bytes, None]:
        """Return the message to be signed."""
        return self.content.chunks()

    def get_private_key(self) -> Any:
        """Return the private key for decryption and signing."""
        return self.spec.get_private_key()

    def get_public_key(self) -> Any:
        """Return the public key for verification and encryption."""
        return self.spec.get_public_key()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import secrets
from typing import cast
from typing import Any

import pydantic

from ckms.utils import current_timestamp
from .audiencetype import AudienceType
from .claimset import ClaimSet
from .invalidtoken import MaximumAgeExceeded
from .invalidtoken import TokenNotEffective
from .invalidtoken import TokenExpired
from .invalidtoken import WrongAudience
from .invalidtoken import WrongIssuer


class JSONWebToken(ClaimSet):
    __module__: str = 'ckms.types'
    _strict_required: set[str] = {'jti', 'exp', 'iat', 'nbf'}
    iss: str | None = None
    sub: str | None = None
    aud: AudienceType = AudienceType()
    exp: int | None = None
    nbf: int | None = None
    iat: int | None = None
    jti: str | None = None

    @classmethod
    def fromjson(cls, payload: bytes | str) -> 'JSONWebToken':
        return cast(cls, super().fromjson(payload))

    @classmethod
    def frompayload(cls, payload: bytes | str) -> 'JSONWebToken':
        return cast(cls, super().frompayload(payload))

    @classmethod
    def strict(
        cls,
        *,
        iss: str,
        ttl: int,
        jti: str | None = None,
        **claims: Any
    ) -> 'JSONWebToken':
        now = current_timestamp()
        exp = now + ttl
        nbf = claims.setdefault('nbf', now)
        if nbf > exp:
            raise ValueError("The 'nbf' claim must not be in the future.")
        return cls.parse_obj({
            **claims,
            'iss': iss,
            'iat': now,
            'exp': exp,
            'jti': jti or secrets.token_urlsafe(24)
        })

    def add_audience(self, audience: str) -> None:
        """Adds an audience to the JSON Web Token (JWT)."""
        self.aud.add(audience)

    def expires_in(
        self,
        ttl: int | None = None,
        now: int | None = None
    ) -> int | None:
        """Set the time-to-live using the ``exp`` claim. Return
        the remaining seconds.
        """
        now = now if now is not None else current_timestamp()
        if ttl is not None:
            self.exp = now + ttl
        return (self.exp - now) if self.exp else None

    def is_selfsigned(self) -> bool:
        """Return a boolean indicating if the token is self-issuer."""
        return self.iss == self.sub and None not in {self.iss, self.sub}

    def verify(
        self,
        *,
        issuer: str | None = None,
        audience: set[str] | None = None,
        required: set[str] | None = None,
        now: int | None = None,
        max_age: int | None = None,
        strict: bool = False
    ) -> None:
        """Verifies the claims according to the specification laid out
        in the JOSE RFCs.
        """
        required = required or set()
        if max_age:
            required.add('iat')
        super().verify(required=required, strict=strict)
        now = now or current_timestamp()
        if self.exp is not None and self.exp <= now:
            raise TokenExpired
        if self.nbf and self.nbf > now:
            raise TokenNotEffective
        if audience and not bool(audience & self.aud):
            raise WrongAudience(allowed=audience)
        if max_age is not None and (now - cast(int, self.iat)) >= max_age:
            assert self.iat is not None
            raise MaximumAgeExceeded(max_age, now, self.iat)
        if issuer is not None and issuer != self.iss:
            raise WrongIssuer


class SecurityEventToken(JSONWebToken):
    """Represents a Security Event Token (SET as defined in :rfc:`8417`."""
    # RFC 8417 Security Event Token (SET)
    iss: str
    iat: int
    jti: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(24),
    )
    aud: AudienceType
    exp: None
    nbf: None
    events: dict[str, dict[str, Any]]
    toe: int | None
    txn: str | None


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`PublicKeySpecification`."""
from typing import Any, Awaitable
from typing import Generator
from ckms.types.decrypter import Decrypter

import pydantic

from ckms.types import CryptographyPublicKeyType
from ckms.types import KeyAlgorithmType
from ckms.types import KeyOperationType
from ckms.types import KeyUseType
from ckms.types import Encrypter
from ckms.types import EdwardsCurveType
from ckms.types import EllipticCurveType
from ckms.types import IKeySpecification
from ckms.types import IProvider
from ckms.types import Signer
from ckms.types import Verifier
from .certificate import Certificate


class PublicKeySpecification(IKeySpecification, Encrypter, Verifier, pydantic.BaseModel):
    private: Decrypter | Signer | None = None
    provider: IProvider
    kid: str
    kty: str
    alg: KeyAlgorithmType
    crv: EdwardsCurveType | EllipticCurveType | None = None
    use: KeyUseType
    key_ops: set[KeyOperationType]
    tags: set[str] = set()
    key: CryptographyPublicKeyType
    certificate: Certificate | None

    @property # type: ignore
    def algorithm(self) -> str:
        return self.alg

    def as_jwk(self, private: bool = False) -> dict[str, Any]:
        assert not private
        claims: dict[str, Any] = {
            **self.provider.inspector.to_jwk(self.key),
            'alg': self.alg,
            'crv': self.crv,
            'use': self.use,
            'key_ops': self.key_ops
        }
        if self.certificate:
            claims.update(self.certificate.claims())
        return claims

    def can_sign(self) -> bool:
        return False

    def can_wrap(self) -> bool:
        """Return a boolean indicating if the key can wrap another key."""
        return KeyOperationType.wrapKey in self.key_ops

    def can_perform(self, op: str) -> bool:
        """Return a boolean indicating if the key can perform the
        given cryptographic operation.
        """
        return op in self.key_ops

    def can_verify(self) -> bool:
        """Return a boolean indicating if this implementation can verify
        a digital signature.
        """
        return self.use == 'sig'

    def decrypt(self, *args: Any, **kwargs: Any) -> bytes | Awaitable[bytes]:
        if self.private is None:
            raise ValueError(f"Private key not available for public key {self.kid}")
        assert isinstance(self.private, Decrypter)
        return self.private.decrypt(*args, **kwargs)

    def get_public_key(self) -> CryptographyPublicKeyType:
        """Return the public key that may be used for encryption or
        verifcation.
        """
        return self.key

    def is_aead(self) -> bool:
        """Return a boolean indicating if the operator supports Authenticated
        Encryption with Associated Data (AEAD).
        """
        return False

    def is_asymmetric(self) -> bool:
        """Return a boolean indicating if the key is an asymmetric
        algorithm.
        """
        return True

    def is_loaded(self) -> bool: 
        """Return a boolean indicating if the metadata describing the
        key is retrieved.
        """
        return True

    def is_public(self) -> bool:
        """Return a boolean indicating if this specification represent
        a public key.
        """
        return True

    def is_symmetric(self) -> bool:
        """Return a boolean indicating if the key is a symmetric
        algorithm.
        """
        return False

    def __await__(self) -> Generator[None, Any, 'PublicKeySpecification']:
        async def f(): return self
        return f().__await__()

    class Config:
        arbitrary_types_allowed: bool = True



"""Declares :class:`OpenAuthorizationClient`."""
from curses import meta
import logging
import typing

import httpx
import pydantic

from .servermetadata import ServerMetadata


class OpenAuthorizationClient(httpx.AsyncClient):
    __module__: str = 'ckms.lib.oauth2'
    metadata: typing.Optional[ServerMetadata] = None
    oauth2_metadata_endpoint: str = "/.well-known/oauth-authorization-server"
    openid_configuration: str = "/.well-known/openid-configuration"
    logger: logging.Logger = logging.getLogger('oauth2.client')
    keys: typing.Dict[str, typing.Any]

    def __init__(
        self,
        server: str,
        logger: typing.Optional[logging.Logger] = None,
        **kwargs: typing.Any
    ):
        kwargs['base_url'] = server
        super().__init__(**kwargs) # type: ignore
        self.logger = logger or self.logger
        self.keys = {}
        self.server = server

    async def get_metadata(self) -> ServerMetadata:
        """Return the server metadata as exposed by the OAuth 2.0
        metadata endpoint or OpenID configuration endpoint.
        """
        try:
            metadata = ServerMetadata(**(
                await self.get_openid_configuration()
                or await self.get_authorization_server_metadata()
                or {}
            ))
            if metadata.jwks_uri:
                self.keys = await self.fetch_metadata(metadata.jwks_uri)
            return metadata
        except pydantic.ValidationError:
            self.logger.warning(
                "Invalid OAuth 2.0 metadata sent by %s", self.server
            )
            return ServerMetadata()

    async def get_authorization_server_metadata(
        self
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        return await self.fetch_metadata(self.oauth2_metadata_endpoint)

    async def get_openid_configuration(
        self
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        return await self.fetch_metadata(self.openid_configuration)

    async def fetch_metadata(
        self,
        url: str
    ) -> typing.Dict[str, typing.Any]:
        try:
            response = await self.get(url) # type: ignore
            return response.json()
        except (httpx.HTTPError, httpx.StreamError, TypeError, ValueError) as exc:
            self.logger.warning(
                "Caught fatal %s during metadata lookup.",
                type(exc).__name__
            )
            return {}

    async def __aenter__(self):
        await super().__aenter__()
        if self.metadata is None:
            self.metadata = await self.get_metadata()
        return self


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import json

import pytest

from ckms.core.models import JOSEObject
from ckms.types import MalformedObject
from ckms.utils import b64decode


def test_deserialize_jwe_json(multi_jwe: bytes):
    obj = JOSEObject.deserialize(multi_jwe)
    assert len(obj['recipients']) == 2
    assert obj['recipients'][0]['header']['foo'] == 1
    assert 'protected' in obj
    assert 'iv' in obj
    assert 'tag' in obj
    assert 'ciphertext' in obj
    assert 'aad' not in obj
    assert 'header' not in obj
    assert 'unprotected' not in obj
    assert 'encrypted_key' not in obj


def test_deserialize_jwe_flattened(flattened_jwe: bytes):
    obj = JOSEObject.deserialize(flattened_jwe)
    protected = json.loads(b64decode(obj['protected']))
    header = obj['recipients'][0]['header']
    assert 'protected' in obj
    assert 'iv' in obj
    assert 'tag' in obj
    assert 'ciphertext' in obj
    assert 'aad' not in obj
    assert 'header' not in obj
    assert 'unprotected' not in obj
    assert 'encrypted_key' not in obj
    assert protected['foo'] == 1
    assert header['bar'] == 1


def test_deserialize_jwe_compact(compact_jwe: bytes):
    obj = JOSEObject.deserialize(compact_jwe)
    protected = json.loads(b64decode(obj['protected']))
    assert 'protected' in obj
    assert 'iv' in obj
    assert 'tag' in obj
    assert 'aad' not in obj
    assert 'ciphertext' in obj
    assert 'header' not in obj
    assert 'unprotected' not in obj
    assert 'encrypted_key' not in obj
    assert 'foo' in protected


def test_deserialize_jws_json(multi_jws: bytes):
    obj = JOSEObject.deserialize(multi_jws)
    assert len(obj['signatures']) == 2
    assert 'protected' in obj['signatures'][0]
    assert 'protected' in obj['signatures'][1]


def test_deserialize_jws_flattened(flattened_jws: bytes):
    obj = JOSEObject.deserialize(flattened_jws)
    header = json.loads(b64decode(obj['signatures'][0]['protected']))
    assert 'payload' in obj
    assert 'signatures' in obj
    assert 'header' not in obj
    assert 'protected' not in obj
    assert header.get('foo') == 1
    assert obj['signatures'][0]['header'].get('bar') == 1


def test_deserialize_jws_compact(compact_jws: bytes):
    obj = JOSEObject.deserialize(compact_jws)
    assert 'payload' in obj
    assert 'signatures' in obj
    assert len(obj['signatures']) == 1


@pytest.mark.parametrize("token", [
    'e', # No segments
    'e.e', # Not enough segments
    'e.e.e.', # Too many segments for signature
    'e.e.e.e.e.e', # Too many segments for encryption,
    'foo', # Invalid base64 encoding
])
def test_malformed_token_raises_malformed_token(token: str):
    with pytest.raises(MalformedObject):
        JOSEObject.deserialize(token)


"""Declares :class:`JSONWebEncryptionBase`."""
import typing

import pydantic

from .jsonb64 import JSONB64
from .octetb64 import OctetB64


class JSONWebEncryptionBase(pydantic.BaseModel):
    __module__: str = 'ckms.core.models'
    protected: JSONB64 = JSONB64()
    unprotected: dict[str, typing.Any] = {}
    iv: OctetB64 | None = None
    aad: OctetB64 | None = None
    ciphertext: OctetB64
    tag: OctetB64 | None

    def get_header(self) -> dict[str, typing.Any]:
        return {
            **self.unprotected,
            **typing.cast(dict[str, typing.Any], self.protected or {}),
        }


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import hashlib

from cryptography.hazmat.primitives import hashes


class AlgorithmIdentifier(str, enum.Enum):
    sha1    = '1.3.14.3.2.26'
    sha256  = '2.16.840.1.101.3.4.2.1'
    sha384  = '2.16.840.1.101.3.4.2.2'
    sha512  = '2.16.840.1.101.3.4.2.3'
    sha224  = '2.16.840.1.101.3.4.2.4'

    @property
    def hasher_class(self) -> type[hashes.Hash]:
        return getattr(hashes, str.upper(self.name))

    def digest(self, data: bytes) -> bytes:
        return getattr(hashlib, self.name)(data).digest()


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from pyasn1_modules import rfc2459


class Attribute(pydantic.BaseModel):
    type: str
    values: list[Any]

    @pydantic.validator('type', pre=True)
    def validate_type(cls, value: str | rfc2459.AttributeType) -> str:
        if isinstance(value, rfc2459.AttributeType):
            value = str(value)
        return value

    @pydantic.validator('values', pre=True)
    def validate_value(cls, value: rfc2459.AttributeValue) -> list[Any]:
        return list(value)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1_modules import rfc5480

from ckms.ext import asn1
from .ellipticcurvetype import EllipticCurveType


class ECParameters(asn1.ASN1Model):
    __asn1spec__: type[rfc5480.ECParameters] = rfc5480.ECParameters
    curve: EllipticCurveType = pydantic.Field(..., alias='namedCurve')

    @pydantic.validator('curve', pre=True)
    def validate_curve(
        cls,
        value: ec.EllipticCurve | str | EllipticCurveType
    ) -> str | EllipticCurveType:
        if isinstance(value, ec.EllipticCurve):
            value = EllipticCurveType.fromcurve(value)
        return value


"""Declares :class:`JSONWebEncryptionFlattened`."""
import typing

from .jsonwebencryptionbase import JSONWebEncryptionBase
from .jsonwebencryptionrecipient import JSONWebEncryptionRecipient
from .octetb64 import OctetB64


class JSONWebEncryptionFlattened(JSONWebEncryptionBase):
    header: dict[str, typing.Any] = {}
    encrypted_key: OctetB64

    @property
    def recipients(self) -> list[JSONWebEncryptionRecipient]:
        return [
            JSONWebEncryptionRecipient(
                header=self.header,
                encrypted_key=OctetB64(self.encrypted_key)
            )
        ]

    def get_header(self) -> dict[str, typing.Any]:
        return {**super().get_header(), **self.header}


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Final

import pytest

import ckms.core
import ckms.core.local
from ckms.core import const
from ckms.core.models import KeySpecification
from ckms.core.provider.tests import AsymmetricSigningTestCase
from ckms.core.provider.tests import EncryptionTestCase
from ckms.core.provider.tests import ProviderTestCase
from ckms.core.provider.tests import SymmetricSigningTestCase


provider: Final = ckms.core.provider('local')

# List of supported keys. For each kty, specify a key that is generated
# by use and a key that is generated by algorithm, appended with a list
# of keys for each algorithm that should be supported by this provider.
SIGNING_KEYS_ASYMMETRIC: list[KeySpecification] = [
    provider.parse_spec({
        'kty': 'EC',
        'use': 'sig',
        'key': {}
    }),
    provider.parse_spec({
        'kty': 'EC',
        'algorithm': const.DEFAULT_EC_SIGNING_ALGORITHM,
        'key': {}
    }),
    provider.parse_spec({
        'kty': 'OKP',
        'use': 'sig',
        'key': {}
    }),
    provider.parse_spec({
        'kty': 'OKP',
        'curve': const.DEFAULT_ED_SIGNING_CURVE,
        'key': {}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'use': 'sig',
        'key': {'length': 1024}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'use': 'sig',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': const.DEFAULT_RSA_SIGNING_ALG,
        'key': {'path': 'pki/rsa.key'}
    }),

    # Explicitly named algorithms
    provider.parse_spec({
        'kty': 'EC',
        'algorithm': 'ES256',
        'key': {'path': 'pki/p256.key'}
    }),
    provider.parse_spec({
        'kty': 'EC',
        'algorithm': 'ES384',
        'key': {'path': 'pki/p384.key'}
    }),
    provider.parse_spec({
        'kty': 'EC',
        'algorithm': 'ES512',
        'key': {'path': 'pki/p521.key'}
    }),
    provider.parse_spec({
        'kty': 'EC',
        'algorithm': 'ES256K',
        'key': {'path': 'pki/p256k.key'}
    }),
    provider.parse_spec({
        'kty': 'OKP',
        'curve': 'Ed448',
        'key': {'path': 'pki/ed448.key'}
    }),
    provider.parse_spec({
        'kty': 'OKP',
        'curve': 'Ed25519',
        'key': {'path': 'pki/ed25519.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RS256',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RS384',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RS512',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'PS256',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'PS384',
        'key': {'path': 'pki/rsa.key'}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'PS512',
        'key': {'path': 'pki/rsa.key'}
    }),
]


SIGNING_KEYS_SYMMETRIC: list[KeySpecification] = [
    provider.parse_spec({
        'kty': 'oct',
        'use': 'sig',
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': const.DEFAULT_SYMMETRIC_SIGNING_ALGORITHM,
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'HS256',
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'HS384',
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'HS512',
        'key': {'length': 32}
    }),
]

ENCRYPTION_KEYS: list[KeySpecification] = [
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A128GCM',
        'key': {'length': 16}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A192GCM',
        'key': {'length': 24}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A256GCM',
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A128GCMKW',
        'key': {'length': 16}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A192GCMKW',
        'key': {'length': 24}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A256GCMKW',
        'key': {'length': 32}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A128KW',
        'key': {'length': 16}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A192KW',
        'key': {'length': 24}
    }),
    provider.parse_spec({
        'kty': 'oct',
        'algorithm': 'A256KW',
        'key': {'length': 32}
    }),
    #provider.parse_spec({
    #    'kty': 'oct',
    #    'algorithm': 'A128CBC-HS256',
    #    'key': {'length': 16}
    #}),
    #provider.parse_spec({
    #    'kty': 'oct',
    #    'algorithm': 'A192CBC-HS384',
    #    'key': {'length': 24}
    #}),
    #provider.parse_spec({
    #    'kty': 'oct',
    #    'algorithm': 'A256CBC-HS512',
    #    'key': {'length': 32}
    #}),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RSA-OAEP',
        'key': {'path': "pki/rsa.key"}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RSA-OAEP-256',
        'key': {'path': "pki/rsa.key"}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RSA-OAEP-384',
        'key': {'path': "pki/rsa.key"}
    }),
    provider.parse_spec({
        'kty': 'RSA',
        'algorithm': 'RSA-OAEP-512',
        'key': {'path': "pki/rsa.key"}
    }),
]


class TestLocalProviderFacilities(ProviderTestCase):
    name: str = 'local'


@pytest.mark.parametrize("spec", ENCRYPTION_KEYS)
class TestEncryption(EncryptionTestCase):
    name: str = 'local'


@pytest.mark.parametrize("spec", SIGNING_KEYS_ASYMMETRIC)
class TestAsymmetricSigning(AsymmetricSigningTestCase):
    name: str = 'local'


@pytest.mark.parametrize("spec", SIGNING_KEYS_SYMMETRIC)
class TestSymmetricSigning(SymmetricSigningTestCase):
    name: str = 'local'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`KeyInspector`."""
import functools
import hashlib
import typing

from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives import serialization

from ckms import types
from ckms.types import IKeyInspector
from ckms.types import KeyOperationType
from ckms.utils import b64encode_int
from ckms.utils import b64encode
from ckms.utils import b64decode_int
from ckms.utils import b64decode


JWKReturnType = dict[str, str | list[str]]
KeyType: typing.TypeAlias = IKeyInspector.KeyType


class KeyInspector(types.IKeyInspector):
    """Inspects keys from the :mod:`cryptography` package."""
    __module__: str = 'ckms.core'

    CurveType: typing.TypeAlias = typing.Union[
        ec.SECP256R1,
        ec.SECP384R1,
        ec.SECP521R1,
        ec.SECP256K1
    ]

    #: Maintains a mapping between curve names.
    curve_mapping: dict[typing.Any, typing.Any] = {
        "P-256": ec.SECP256R1,
        "P-384": ec.SECP384R1,
        "P-521": ec.SECP521R1,
        "P-256K": ec.SECP256K1,
        types.EllipticCurveType.P256: ec.SECP256R1,
        types.EllipticCurveType.P384: ec.SECP384R1,
        types.EllipticCurveType.P521: ec.SECP521R1,
        types.EllipticCurveType.P256K: ec.SECP256K1,
        ec.SECP256R1.name: "P-256",
        ec.SECP384R1.name: "P-384",
        ec.SECP521R1.name: "P-521",
        ec.SECP256K1.name: "P-256K"
    }

    #: Maps types to curves.
    ed_curve_map: dict[typing.Any, typing.Any] = {
        ed448.Ed448PublicKey: "Ed448",
    }
    ed_curve_map.update({y: x for x, y in ed_curve_map.items()})

    #: Map algorithms to curves
    algorithm_curves: dict[str, str] = {
        'ES256': 'P-256',
        'ES384': 'P-384',
        'ES512': 'P-521',
        'ES256K': 'P-256K',
    }

    #: Map algorithms to digest algorithms
    algorithm_digests: dict[str, str] = {
        'ES256'         : 'sha256',
        'ES384'         : 'sha384',
        'ES512'         : 'sha512',
        'ES256K'        : 'sha256',
        'HS256'         : 'sha256',
        'HS384'         : 'sha384',
        'HS512'         : 'sha512',
        'PS256'         : 'sha256',
        'PS384'         : 'sha384',
        'PS512'         : 'sha512',
        'RS256'         : 'sha256',
        'RS384'         : 'sha384',
        'RS512'         : 'sha512',
    }

    #: Maps algorithms to uses
    algorithm_use: dict[str, str] = {
        'EdDSA'         : 'sig',
        'ES256'         : 'sig',
        'ES384'         : 'sig',
        'ES512'         : 'sig',
        'ES256K'        : 'sig',
        'HS256'         : 'sig',
        'HS384'         : 'sig',
        'HS512'         : 'sig',
        'PS256'         : 'sig',
        'PS384'         : 'sig',
        'PS512'         : 'sig',
        'RS256'         : 'sig',
        'RS384'         : 'sig',
        'RS512'         : 'sig',
        'ECDH-ES'       : 'enc',
        'ECDH-ES+A128KW': 'enc',
        'ECDH-ES+A192KW': 'enc',
        'ECDH-ES+A256KW': 'enc',
        'RSA1_5'        : 'enc',
        'RSA-OAEP'      : 'enc',
        'RSA-OAEP-256'  : 'enc',
        'RSA-OAEP-384'  : 'enc',
        'RSA-OAEP-512'  : 'enc',
        'A128KW'        : 'enc',
        'A192KW'        : 'enc',
        'A256KW'        : 'enc',
        'A128GCM'       : 'enc',
        'A192GCM'       : 'enc',
        'A256GCM'       : 'enc',
        'A128GCMKW'     : 'enc',
        'A192GCMKW'     : 'enc',
        'A256GCMKW'     : 'enc',
    }

    @staticmethod
    def get_ed_curve(
        key: (
            ed448.Ed448PublicKey
            | ed25519.Ed25519PublicKey
            | x448.X448PublicKey
            | x25519.X25519PublicKey
        )
    ) -> str:
        """Return the name of the curve."""
        if isinstance(key, (ed448.Ed448PrivateKey, ed448.Ed448PublicKey)):
            return "Ed448"
        elif isinstance(key, (x448.X448PrivateKey, x448.X448PublicKey)):
            return "X448"
        elif isinstance(key, (ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey)):
            return "Ed25519"
        elif isinstance(key, (x25519.X25519PrivateKey, x25519.X25519PublicKey)): #type: ignore
            return "X25519"
        else:
            raise NotImplementedError

    @functools.singledispatchmethod
    def calculate_kid(
        self,
        key: KeyType | algorithms.AES | str
    ) -> str:
        """Calculate a key identifier for the given `key`.
        
        If `key` is a string, it is assumed to be the resource address
        of a cloud KMS resource, otherwise `key` must be a key type from
        the :mod:`cryptography` package.
        """
        raise NotImplementedError(key)

    @calculate_kid.register
    def _calculate_kid_ec_public(self, key: ec.EllipticCurvePublicKey) -> str:
        h = hashlib.sha3_256() # nosec
        l = (key.curve.key_size + 7) // 8
        n = key.public_numbers()
        h.update(str.encode(self.curve_mapping[key.curve.name]))
        h.update(str.encode('EC'))
        h.update(int.to_bytes(n.x, l, 'big'))
        h.update(int.to_bytes(n.y, l, 'big'))
        return self.encode_kid(h.digest())

    @calculate_kid.register
    def _calculate_kid_ec_private(self, key: ec.EllipticCurvePrivateKey) -> str:
        return self._calculate_kid_ec_public(key.public_key())

    @calculate_kid.register
    def _calculate_kid_ed448_public(self, key: ed448.Ed448PublicKey) -> str:
        return self._calculate_kid_ed(key)

    @calculate_kid.register
    def _calculate_kid_ed448_private(self, key: ed448.Ed448PrivateKey) -> str:
        return self._calculate_kid_ed(key.public_key())

    @calculate_kid.register
    def _calculate_kid_ed25519_public(self, key: ed25519.Ed25519PublicKey) -> str:
        return self._calculate_kid_ed(key)

    @calculate_kid.register
    def _calculate_kid_ed25519_private(self, key: ed25519.Ed25519PrivateKey) -> str:
        return self._calculate_kid_ed(key.public_key())

    @calculate_kid.register
    def _calculate_kid_x448_public(self, key: x448.X448PublicKey) -> str:
        return self._calculate_kid_ed(key)

    @calculate_kid.register
    def _calculate_kid_x448_private(self, key: x448.X448PrivateKey) -> str:
        return self._calculate_kid_ed(key.public_key())

    @calculate_kid.register
    def _calculate_kid_x25519_public(self, key: x25519.X25519PublicKey) -> str:
        return self._calculate_kid_ed(key)

    @calculate_kid.register
    def _calculate_kid_x25519_private(self, key: x25519.X25519PrivateKey) -> str:
        return self._calculate_kid_ed(key.public_key())

    def _calculate_kid_ed(
        self,
        key: (
            ed448.Ed448PublicKey
            | ed25519.Ed25519PublicKey
            | x448.X448PublicKey
            | x25519.X25519PublicKey
        )
    ) -> str:
        h = hashlib.sha3_256()
        h.update(str.encode(self.get_ed_curve(key)))
        h.update(str.encode("OKP"))
        h.update(
            key.public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw
            )
        )
        return self.encode_kid(h.digest())

    @calculate_kid.register
    def _calculate_kid_rsa_public(self, key: rsa.RSAPublicKey) -> str:
        h = hashlib.sha3_256()
        l = key.key_size // 8
        n = key.public_numbers()
        h.update(int.to_bytes(n.e, 3, 'big'))
        h.update(str.encode("RSA"))
        h.update(int.to_bytes(n.n, l, 'big'))
        return self.encode_kid(h.digest())

    @calculate_kid.register
    def _calculate_kid_rsa_private(self, key: rsa.RSAPrivateKey) -> str:
        return self._calculate_kid_rsa_public(key.public_key())

    @calculate_kid.register
    def _calculate_kid_aes(self, algorithm: algorithms.AES) -> str:
        return self._calculate_kid_bytes(algorithm.key)

    @calculate_kid.register
    def _calculate_kid_str(self, key: str) -> str:
        h = hashlib.sha3_256() # nosec
        h.update(str.encode(key, 'ascii'))
        h.update(b'kty')
        return self.encode_kid(h.digest())

    @calculate_kid.register
    def _calculate_kid_bytes(self, value: bytes) -> str:
        # Generic method to calculate a key identifier based
        # on an opaque external identifier, such as the address
        # of a cloud KMS resource.
        h = hashlib.sha3_256(value)
        h.update(str.encode("oct"))
        return self.encode_kid(h.digest())

    def to_bytes(self, key: typing.Any) -> bytes:
        # TODO: Not all keys use the PEM format.
        return self.to_pem(key)

    @functools.singledispatchmethod
    def to_pem(self, key: KeyType) -> bytes:
        # TODO: HMAC/AES are not in the PEM format.
        if isinstance(key, bytes):
            return key
        elif isinstance(key, algorithms.AES):
            return key.key
        else:
            raise NotImplementedError(type(key))

    @to_pem.register
    def _ed448_public_to_pem(self, key: ed448.Ed448PublicKey) -> bytes:
        return self._ed_public_to_pem(key)

    @to_pem.register
    def _ed448_private_to_pem(self, key: ed448.Ed448PrivateKey) -> bytes:
        return self._ed_private_to_pem(key)

    @to_pem.register
    def _x448_public_to_pem(self, key: x448.X448PublicKey) -> bytes:
        return self._ed_public_to_pem(key)

    @to_pem.register
    def _x448_private_to_pem(self, key: x448.X448PrivateKey) -> bytes:
        return self._ed_private_to_pem(key)

    @to_pem.register
    def _ed25519_public_to_pem(self, key: ed25519.Ed25519PublicKey) -> bytes:
        return self._ed_public_to_pem(key)

    @to_pem.register
    def _ed25519_private_to_pem(self, key: ed25519.Ed25519PrivateKey) -> bytes:
        return self._ed_private_to_pem(key)

    @to_pem.register
    def _x25519_public_to_pem(self, key: x25519.X25519PublicKey) -> bytes:
        return self._ed_public_to_pem(key)

    @to_pem.register
    def _x25519_private_to_pem(self, key: x25519.X25519PrivateKey) -> bytes:
        return self._ed_private_to_pem(key)

    def _ed_public_to_pem(
        self,
        key: (
            ed448.Ed448PublicKey
            | ed25519.Ed25519PublicKey
            | x448.X448PublicKey
            | x25519.X25519PublicKey
        )
    ) -> bytes:
        return key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )

    def _ed_private_to_pem(
        self,
        key: (
            ed448.Ed448PrivateKey
            | ed25519.Ed25519PrivateKey
            | x448.X448PrivateKey
            | x25519.X25519PrivateKey
        )
    ) -> bytes:
        return key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        )

    @to_pem.register
    def _ec_public_to_pem(self, key: ec.EllipticCurvePublicKey) -> bytes:
        return self._pkcs_public_to_pem(key)

    @to_pem.register
    def _ec_private_to_pem(self, key: ec.EllipticCurvePrivateKey) -> bytes:
        return self._pkcs_private_to_pem(key)

    @to_pem.register
    def _rsa_public_to_pem(self, key: rsa.RSAPublicKey) -> bytes:
        return self._pkcs_public_to_pem(key)

    @to_pem.register
    def _rsa_private_to_pem(self, key: rsa.RSAPrivateKey) -> bytes:
        return self._pkcs_private_to_pem(key)

    def _pkcs_public_to_pem(self, key: rsa.RSAPublicKey|ec.EllipticCurvePublicKey):
        return key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    def _pkcs_private_to_pem(self, key: rsa.RSAPrivateKey|ec.EllipticCurvePrivateKey):
        return key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )

    def get_algorithm_use(self, algorithm: str) -> str:
        """Return the `use` parameter based on a given algorithm."""
        return self.algorithm_use[algorithm]

    @functools.singledispatchmethod
    def to_jwk(
        self,
        key: KeyType | algorithms.AES | bytes
    ) -> JWKReturnType:
        """Return a dictionary holding a representation of the key as
        a JSON Web Key (JWK).
        """
        raise NotImplementedError

    @to_jwk.register
    def _aes_to_jwk(self, algorithm: algorithms.AES) -> JWKReturnType:
        return self._jwk_properties(
            key=algorithm,
            kty="oct",
            k=bytes.decode(b64encode(algorithm.key), 'ascii')
        )

    @to_jwk.register
    def _bytes_to_jwk(self, key: bytes) -> JWKReturnType:
        return self._jwk_properties(
            key=key,
            kty="oct",
            k=bytes.decode(b64encode(key), 'ascii')
        )

    @to_jwk.register
    def _ec_public_to_jwk(self, key: ec.EllipticCurvePublicKey) -> JWKReturnType:
        numbers = key.public_numbers()
        return self._jwk_properties(
            key=key,
            kty="EC",
            crv=self.curve_mapping[key.curve.name],
            x=b64encode_int(numbers.x),
            y=b64encode_int(numbers.y)
        )

    @to_jwk.register
    def _ec_private_to_jwk(self, key: ec.EllipticCurvePrivateKey) -> JWKReturnType:
        numbers = key.private_numbers()
        return self._jwk_properties(
            key=key,
            kty="EC",
            crv=self.curve_mapping[key.curve.name],
            x=b64encode_int(numbers.public_numbers.x),
            y=b64encode_int(numbers.public_numbers.y),
            d=b64encode_int(numbers.private_value)
        )

    @to_jwk.register
    def _ed448_public_to_jwk(self, key: ed448.Ed448PublicKey) -> JWKReturnType:
        return self._jwk_properties(
            key=key,
            kty="OKP",
            alg='EdDSA',
            use='sig',
            crv="Ed448",
            x=b64encode(
                key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            )
        )

    @to_jwk.register
    def _ed448_private_to_jwk(self, key: ed448.Ed448PrivateKey) -> JWKReturnType:
        public = key.public_key()
        return self._jwk_properties(
            key=key,
            kty="OKP",
            alg='EdDSA',
            use='sig',
            crv="Ed448",
            x=b64encode(
                public.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            ),
            d=b64encode(
                key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption()
                )
            )
        )

    @to_jwk.register
    def _x448_public_to_jwk(self, key: x448.X448PublicKey) -> JWKReturnType:
        return self._jwk_properties(
            key=key,
            kty="OKP",
            use='enc',
            crv="X448",
            x=b64encode(
                key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            )
        )

    @to_jwk.register
    def _x448_private_to_jwk(self, key: x448.X448PrivateKey) -> JWKReturnType:
        public = key.public_key()
        return self._jwk_properties(
            key=key,
            kty="OKP",
            use='enc',
            crv="X448",
            x=b64encode(
                public.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            ),
            d=b64encode(
                key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption()
                )
            )
        )

    @to_jwk.register
    def _ed25519_public_to_jwk(self, key: ed25519.Ed25519PublicKey) -> JWKReturnType:
        return self._jwk_properties(
            key=key,
            kty="OKP",
            alg='EdDSA',
            use='sig',
            crv="Ed25519",
            x=b64encode(
                key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            )
        )

    @to_jwk.register
    def _ed25519_private_to_jwk(self, key: ed25519.Ed25519PrivateKey) -> JWKReturnType:
        public = key.public_key()
        return self._jwk_properties(
            key=key,
            kty="OKP",
            alg='EdDSA',
            use='sig',
            crv="Ed25519",
            x=b64encode(
                public.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            ),
            d=b64encode(
                key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption()
                )
            )
        )

    @to_jwk.register
    def _x25519_public_to_jwk(self, key: x25519.X25519PublicKey) -> JWKReturnType:
        return self._jwk_properties(
            key=key,
            kty="OKP",
            use="enc",
            crv="X25519",
            x=b64encode(
                key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            )
        )

    @to_jwk.register
    def _x25519_private_to_jwk(self, key: x25519.X25519PrivateKey) -> JWKReturnType:
        public = key.public_key()
        return self._jwk_properties(
            key=key,
            kty="OKP",
            use="enc",
            crv="X25519",
            x=b64encode(
                public.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw
                )
            ),
            d=b64encode(
                key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption()
                )
            )
        )

    @to_jwk.register
    def _rsa_public_to_jwk(self, key: rsa.RSAPublicKey) -> JWKReturnType:
        numbers = key.public_numbers()
        return self._jwk_properties(
            key=key,
            kty="RSA",
            n=b64encode_int(numbers.n),
            e=b64encode_int(numbers.e)
        )

    @to_jwk.register
    def _rsa_private_to_jwk(self, key: rsa.RSAPrivateKey) -> JWKReturnType:
        numbers = key.private_numbers()
        return self._jwk_properties(
            key=key,
            kty="RSA",
            n=b64encode_int(numbers.public_numbers.n),
            e=b64encode_int(numbers.public_numbers.e),
            d=b64encode_int(numbers.d),
            p=b64encode_int(numbers.p),
            q=b64encode_int(numbers.q),
            dp=b64encode_int(numbers.dmp1),
            dq=b64encode_int(numbers.dmq1),
            qi=b64encode_int(numbers.iqmp),
        )

    def _jwk_properties(
        self,
        key: KeyType | algorithms.AES | bytes,
        **claims: typing.Any
    ) -> JWKReturnType:
        return {
            **claims,
            'kid': self.calculate_kid(key)
        }

    def from_jwk(
        self,
        jwk: dict[str, str]
    ) -> KeyType | bytes: # pragma: no cover
        """Create a new key instance from a dictionary holding a JSON
        Web Key (JWK).
        """
        kty = jwk.get('kty')
        crv = jwk.get('crv')
        claims = set(jwk.keys())
        rsa_claims = {'d', 'p', 'q', 'dp', 'dq', 'qi'}
        if kty is None:
            raise ValueError("The 'kty' claim needs to be defined.")
        if kty not in {"oct", "EC", "OKP", "RSA"}:
            raise ValueError(f"Unknown key type: '{kty}'")
        if (kty == "EC" and crv not in self.curve_mapping)\
        or kty == "OKP" and crv not in {"Ed448", "X448", "Ed25519", "X25519"}:
            raise ValueError("The 'crv' claim must contain a known curve.")
        if kty == "oct":
            if 'k' not in claims:
                raise ValueError("Octet keys need the 'k' claim.")
        elif kty == "EC":
            if not claims > {'x', 'y'}:
                raise ValueError(
                    "Elliptic curve keys need the 'x', 'y' "
                    "and optionally 'd' claims."
                )
        elif kty == "OKP":
            if 'x' not in claims:
                raise ValueError(
                    "EdDSA keys need the 'x' and optionally the 'd' "
                    "claims."
                )
        elif kty == "RSA":
            if not claims >= {'e', 'n'}:
                raise ValueError(
                    "RSA keys need at least the 'e' and 'n' "
                    "parameters."
                )
            if (rsa_claims & claims) and not claims > rsa_claims:
                raise ValueError("Missing claims for RSA private key.")

        key = None
        if kty == "oct":
            key = b64decode(jwk['k'])
        elif kty == "EC":
            numbers = ec.EllipticCurvePublicNumbers(
                curve=self.curve_mapping[crv](),
                x=b64decode_int(jwk['x']),
                y=b64decode_int(jwk['y']),
            )
            key = numbers.public_key()
            if 'd' in jwk:
                numbers = ec.EllipticCurvePrivateNumbers(
                    private_value=b64decode_int(jwk['d']),
                    public_numbers=numbers
                )
                key = numbers.private_key()

        elif kty == "OKP":
            public_factory = private_factory = None
            if crv == "Ed448":
                public_factory, private_factory = (
                    ed448.Ed448PublicKey.from_public_bytes,
                    ed448.Ed448PrivateKey.from_private_bytes
                )
            elif crv == "X448":
                public_factory, private_factory = (
                    x448.X448PublicKey.from_public_bytes,
                    x448.X448PrivateKey.from_private_bytes
                )
            elif crv == "Ed25519":
                public_factory, private_factory = (
                    ed25519.Ed25519PublicKey.from_public_bytes,
                    ed25519.Ed25519PrivateKey.from_private_bytes
                )
            elif crv == "X25519":
                public_factory, private_factory = (
                    x25519.X25519PublicKey.from_public_bytes,
                    x25519.X25519PrivateKey.from_private_bytes
                )
            assert public_factory is not None # nosec
            assert private_factory is not None # nosec
            key = public_factory(b64decode(jwk['x']))
            if 'd' in claims:
                key = private_factory(b64decode(jwk['d']))

        elif kty == "RSA":
            numbers = rsa.RSAPublicNumbers(
                e=b64decode_int(jwk['e']),
                n=b64decode_int(jwk['n']),
            )
            key = numbers.public_key()
            if rsa_claims & claims:
                numbers = rsa.RSAPrivateNumbers(
                    public_numbers=numbers,
                    d=b64decode_int(jwk['d']),
                    p=b64decode_int(jwk['p']),
                    q=b64decode_int(jwk['q']),
                    dmp1=b64decode_int(jwk['dp']),
                    dmq1=b64decode_int(jwk['dq']),
                    iqmp=b64decode_int(jwk['qi'])
                )
                key = numbers.private_key()
        else:
            raise ValueError("Unable to parse the JWK.")
        assert key is not None # nosec
        return key

    def encode_kid(self, digest: bytes) -> str:
        return bytes.decode(b64encode(digest), 'ascii') # nosec

    def get_algorithm_curve(self, algorithm: str) -> str | None:
        return self.algorithm_curves.get(algorithm)

    def get_ec_class(
        self,
        curve: types.EllipticCurveType | str
    ) -> tuple[typing.Callable[..., ec.EllipticCurvePrivateKey], ec.EllipticCurve]:
        return ec.generate_private_key, self.curve_mapping[curve]

    def get_ed_class(
        self,
        curve: str
    ) -> ed448.Ed448PrivateKey | ed25519.Ed25519PrivateKey | x448.X448PrivateKey | x25519.X25519PrivateKey:
        if curve not in {'Ed448', 'Ed25519', 'X448', 'X25519'}: # pragma: no cover
            # Should not happen, but check anyway as a failsafe against RCE
            raise ValueError("Unsupported curve.")
        return getattr(globals()[str.lower(curve)], f'{curve}PrivateKey')

    def get_hashing_algorithm(self, algorithm: str) -> str | None:
        """Return a string indicating the hashing algorithm."""
        return self.algorithm_digests.get(algorithm)

    def get_public_key_ops(self, private: list[KeyOperationType]) -> list[KeyOperationType]:
        public: list[KeyOperationType] = []
        if KeyOperationType.decrypt in private:
            public.append(KeyOperationType.encrypt)
        if KeyOperationType.deriveKey in private:
            public.append(KeyOperationType.deriveKey)
        if KeyOperationType.sign in private:
            public.append(KeyOperationType.verify)
        if KeyOperationType.unwrapKey in private:
            public.append(KeyOperationType.wrapKey)
        return public


"""Declares :class:`JSONWebEncryptionWithRecipients`."""
import typing

from .jsonwebencryptionbase import JSONWebEncryptionBase
from .jsonwebencryptionrecipient import JSONWebEncryptionRecipient


class JSONWebEncryptionWithRecipients(JSONWebEncryptionBase):
    __module__: str = 'ckms.core.models'
    recipients: list[JSONWebEncryptionRecipient]

    def get_header(self) -> dict[str, typing.Any]:
        return {**self.protected, **self.unprotected}


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Data`."""
import hashlib
import io
import pathlib
from typing import Any
from typing import AsyncGenerator
from typing import Callable
from typing import Generator
from typing import TypeAlias

import aiofiles
from aiofiles.threadpool.text import AsyncTextIOWrapper
from aiofiles.threadpool.binary import AsyncBufferedReader

from ..utils import b64encode
from .ihasher import IHasher


File: TypeAlias = AsyncBufferedReader | AsyncTextIOWrapper


class Data:
    """A base class for objects on which a cryptographic operation may
    be performed, such as signing or encryption.
    """
    __module__: str = 'ckms.types'
    _buf: bytes | None
    _algorithm: str | None
    _src: pathlib.Path | None
    content_type: str | None
    encoders: dict[str, Callable[..., bytes]] = {
        'base64': b64encode
    }
    hashers: dict[str, Callable[..., IHasher]] = { # type: ignore
        'sha256': hashlib.sha256,
        'sha384': hashlib.sha384,
        'sha512': hashlib.sha512,
        'sha3_256': hashlib.sha3_256,
        'sha3_384': hashlib.sha3_384,
        'sha3_512': hashlib.sha3_512,
    }
    readlen: int

    @property
    def digestmod(self) -> str | None:
        return self._algorithm

    def __init__(
        self,
        buf: bytes | None = None,
        src: str | pathlib.Path | None = None,
        readlen: int = 1048,
        digest: str | None = None,
        content_type: str | None = None
    ):
        self._algorithm = digest
        if not ((buf is not None) ^ (src is not None)):
            raise TypeError("Provide either the `buf` or `src` parameter.")
        self._buf = buf
        if src is not None and not isinstance(src, pathlib.Path):
            src = pathlib.Path(src)
        self._src = src
        self.content_type = content_type
        self.readlen = readlen

    def base64(self) -> bytes:
        return b64encode(bytes(self))

    async def as_bytes(self) -> bytes:
        if self._buf is None:
            self._buf = b''
            async for chunk in self.chunks():
                self._buf += chunk
        return self._buf

    async def digest(
        self,
        algorithm: str | None = None,
        length: int | None = None,
        encoding: str | None = None
    ) -> bytes:
        """Return a digest of the data using the given digest algorithm."""
        algorithm = algorithm or self._algorithm
        if algorithm is None:
            raise TypeError("The `algorithm` parameter can not be None.")
        hasher = self.hashers[algorithm]()
        async for chunk in self.chunks(length=length):
            hasher.update(chunk)
        return self.encode(hasher.digest(), encoding=encoding)

    async def _read(self) -> 'Data':
        await self.as_bytes()
        return self

    def encode(self, data: bytes, encoding: str | None) -> bytes:
        """Encode the data in the given encoding."""
        if encoding is not None:
            data = self.encoders[encoding](data)
        return data

    async def chunks(self, length: int | None = None) -> AsyncGenerator[bytes, None]:
        length = length or self.readlen
        if self._buf is not None:
            buf = io.BytesIO(self._buf)
            chunk = buf.read(length)
            while chunk:
                yield chunk
                chunk = buf.read(length)
            return
        async with aiofiles.open(self._src, 'rb') as f: # type: ignore
            self._buf = b''
            chunk = await f.read(length) # type: ignore
            while chunk:
                self._buf += chunk
                yield chunk
                chunk = await f.read(length) # type: ignore

    def set_digest_algorithm(self, algorithm: str) -> None:
        self._algorithm = algorithm

    def __bytes__(self) -> bytes:
        if self._buf is not None: return self._buf
        with open(self._src, 'rb') as f: # type: ignore
            return f.read() # type: ignore

    def __await__(self) -> Generator[Any, None, 'Data']:
        return self._read().__await__()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Provider`."""
import functools
from typing import Any
from typing import Awaitable
from ckms.types.ciphertext import CipherText

from cryptography.hazmat.primitives.keywrap import aes_key_wrap

from ckms import types
from ckms.types import AESOperation
from ckms.types import AESKeywrapOperation
from ckms.types import EdDSAOperation
from ckms.types import EllipticCurveOperation
from ckms.types import HMACOperation
from ckms.types import IProvider
from ckms.types import JSONWebKey
from ckms.types import KeywrapOperation
from ckms.types import RSAOperation
from ckms.types import SigningOperation
from ckms.core.models import KeySpecification
from ckms.utils import b64encode
from ..keyinspector import KeyInspector


class Provider(IProvider):
    __module__: str = 'ckms.core'
    algorithm_operations: dict[str, Any] = {
        'A128KW'        : AESKeywrapOperation,
        'A192KW'        : AESKeywrapOperation,
        'A256KW'        : AESKeywrapOperation,
        'A128GCM'       : AESOperation,
        'A192GCM'       : AESOperation,
        'A256GCM'       : AESOperation,
        'A128GCMKW'     : AESOperation,
        'A192GCMKW'     : AESOperation,
        'A256GCMKW'     : AESOperation,
        'EdDSA'         : EdDSAOperation,
        'ES256'         : EllipticCurveOperation,
        'ES384'         : EllipticCurveOperation,
        'ES512'         : EllipticCurveOperation,
        'ES256K'        : EllipticCurveOperation,
        'HS256'         : HMACOperation,
        'HS384'         : HMACOperation,
        'HS512'         : HMACOperation,
        'PS256'         : RSAOperation,
        'PS384'         : RSAOperation,
        'PS512'         : RSAOperation,
        'RS256'         : RSAOperation,
        'RS384'         : RSAOperation,
        'RS512'         : RSAOperation,
        'RSA1_5'        : RSAOperation,
        'RSA-OAEP'      : RSAOperation,
        'RSA-OAEP-256'  : RSAOperation,
        'RSA-OAEP-384'  : RSAOperation,
        'RSA-OAEP-512'  : RSAOperation,
    }
    inspector: KeyInspector
    name: str
    spec: type[KeySpecification] | None = None

    def __init__(
        self,
        inspector: KeyInspector | None = None
    ):
        self._providers = Provider._providers
        self.inspector = inspector or KeyInspector()
        self.on_init()

    def jwk(self, spec: KeySpecification, private: bool, **claims: Any) -> JSONWebKey:
        key = spec.get_public_key() if not private else spec.get_private_key()
        ops = spec.key_ops
        if not private:
            ops = spec.as_public().key_ops
        return JSONWebKey.parse_obj({
            **self.inspector.to_jwk(key),
            'kid': spec.kid,
            'kty': spec.kty,
            'alg': spec.algorithm,
            'key_ops': ops,
            **claims
        })

    def on_init(self) -> None:
        """Called when the :class:`Provider` is initialized."""
        pass

    def configure(self, **params: Any) -> None:
        """Configures the provider with the given parameters."""
        raise NotImplementedError

    def get_hashing_algorithm(self, algorithm: str) -> str | None:
        """Return a string indicating the hashing algorithm."""
        return self.inspector.get_hashing_algorithm(algorithm)

    def parse_spec(self, spec: dict[str, Any]) -> KeySpecification:
        """Parse a key specification from a dictionary."""
        assert self.spec is not None # nosec
        return self.spec.parse_obj({
            **spec,
            'provider': self
        })

    async def fetch(self, blob: Any) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        raise NotImplementedError

    async def random(self, length: int) -> bytes:
        """Return a random byte-sequence of the given length."""
        random = self.get('random')
        return await random.random(length=length)

    async def random_urlsafe(self, length: int) -> str:
        """Like :meth:`random()`, but returs a Base64, URL-safe encoded
        string.
        """
        return bytes.decode(b64encode(await self.random(length)))

    def calculate_kid(self, value: Any) -> str:
        return self.inspector.calculate_kid(value)

    def decrypt(
        self,
        key: KeySpecification,
        algorithm: str,
        ciphertext: types.CipherText
    ) -> bytes | Awaitable[bytes]:
        op = self.algorithm_operations[algorithm].fromalgorithm(
            algorithm=algorithm,
            spec=key,
            content=ciphertext,
            aad=ciphertext.aad,
            iv=ciphertext.iv,
            tag=ciphertext.tag
        )
        return self._decrypt(op)

    @functools.singledispatchmethod
    def _decrypt(self, op: types.Operation) -> bytes | Awaitable[bytes]:
        raise NotImplementedError(
            f"Operation {type(op).__name__} is not implemented by "
            f"{type(self).__name__}."
        )

    @_decrypt.register
    def _decrypt_aes(self, op: types.AESOperation) -> bytes | Awaitable[bytes]:
        return self.decrypt_aes(op)

    @_decrypt.register
    def _decrypt_aes_wrap(self, op: types.AESKeywrapOperation) -> bytes | Awaitable[bytes]:
        return self._unwrap(op)

    def decrypt_aes(self, op: types.AESOperation) -> bytes | Awaitable[bytes]:
        """Decrypt a plain text using Advances Encryption Standard (AES)."""
        raise NotImplementedError(
            f"AES decryption is not supported by {type(self).__name__}"
        )

    @_decrypt.register
    def _decrypt_rsa(self, op: types.RSAOperation) -> bytes | Awaitable[bytes]:
        return self.decrypt_rsa(op)

    def decrypt_rsa(self, op: types.RSAOperation) -> bytes | Awaitable[bytes]:
        """Decrypt a plain text using RSA."""
        raise NotImplementedError(
            f"RSA decryption is not supported by {type(self).__name__}"
        )

    def encrypt(
        self,
        key: KeySpecification,
        algorithm: str,
        plaintext: types.PlainText,
        **kwargs: Any
    ) -> types.EncryptResult:
        """Encrypt the given message with the specified key."""
        op = self.algorithm_operations[algorithm].fromalgorithm(
            algorithm=algorithm,
            spec=key,
            content=plaintext,
            **kwargs
        )
        return self._encrypt(op)

    @functools.singledispatchmethod # type: ignore
    def _encrypt(
        self,
        op: types.Operation
    ) -> types.EncryptResult:
        raise NotImplementedError(
            f"Operation {type(op).__name__} is not implemented by "
            f"{type(self).__name__}."
        )

    @_encrypt.register
    def _encrypt_aes(self, op: types.AESOperation) -> types.EncryptResult:
        return self.encrypt_aes(op)

    @_encrypt.register
    def _encrypt_aes_wrap(self, op: types.AESKeywrapOperation) -> types.EncryptResult:
        return self._wrap(op)

    @_encrypt.register
    def _encrypt_rsa(self, op: types.RSAOperation) -> types.EncryptResult:
        return self.encrypt_rsa(op)

    def encrypt_aes(self, op: types.AESOperation) -> types.EncryptResult:
        """Encrypt a plain text using Advanced Encryption Standard (AES)."""
        raise NotImplementedError(
            f"AES encryption is not supported by {type(self).__name__}"
        )

    def encrypt_rsa(self, op: types.RSAOperation) -> types.EncryptResult:
        """Encrypt a plain text using RSA."""
        raise NotImplementedError(
            f"RSA encryption is not supported by {type(self).__name__}"
        )

    def unwrap(self, op: types.KeywrapOperation) -> bytes | Awaitable[bytes]:
        """Wrap a key using a symmetric encryption algorithm."""
        raise NotImplementedError(
            f"Key unwrapping is not supported by {type(self).__name__}"
        )

    @functools.singledispatchmethod
    def _unwrap(self, op: types.KeywrapOperation) -> bytes | Awaitable[bytes]:
        """Encrypt a plain text using RSA."""
        raise NotImplementedError(
            f"Key unwrapping is not supported by {type(self).__name__}"
        )

    @_unwrap.register
    def _unwrap_aes_wrap(self, op: types.AESKeywrapOperation) -> bytes | Awaitable[bytes]:
        return self.unwrap_aes_wrap(op)

    def unwrap_aes_wrap(self, op: types.AESKeywrapOperation) -> bytes | Awaitable[bytes]:
        """Unwrap a key using AES Key Wrapping."""
        raise NotImplementedError(
            f"AES key unwrapping is not supported by {type(self).__name__}"
        )

    def wrap(
        self,
        key: Any,
        algorithm: str,
        wrap: types.PlainText,
        **kwargs: Any
    ) -> Any | Awaitable[Any]:
        """Wrap key `wrap` using the given `key`."""
        op = self.algorithm_operations[algorithm].fromalgorithm(
            algorithm=algorithm,
            spec=key,
            content=wrap,
            **kwargs
        )
        func = self._wrap
        if not isinstance(op, (KeywrapOperation, AESKeywrapOperation)):
            func = self._encrypt
        return func(op)

    @functools.singledispatchmethod
    def _wrap(self, op: KeywrapOperation) -> types.EncryptResult:
        raise NotImplementedError(
            f"Operation {type(op).__name__} is not implemented by "
            f"{type(self).__name__}."
        )

    @_wrap.register
    def _wrap_aeskw_(self, op: AESKeywrapOperation) -> types.EncryptResult:
        spec = op.get_keyspec()
        return CipherText(
            buf=aes_key_wrap(
                wrapping_key=spec.get_private_bytes(),
                key_to_wrap=bytes(op.content)
            )
        )

    def sign(
        self,
        key: KeySpecification,
        algorithm: str,
        message: types.Digest | types.Message
    ) -> bytes | Awaitable[bytes]:
        """Sign the given digest or message with the specified key. Return
        a byte-sequence holding the signature.
        """
        op = self.algorithm_operations[algorithm].fromalgorithm(
            algorithm=algorithm,
            spec=key,
            content=message
        )
        return self._sign(op)

    @functools.singledispatchmethod # type: ignore
    def _sign(self, op: types.SigningOperation) -> IProvider.SignResult: # pragma: no cover
        raise NotImplementedError(
            f"Operation {type(op).__name__} is not implemented by "
            f"{type(self).__name__}."
        )

    @_sign.register
    def _sign_ec(self, op: EllipticCurveOperation) -> IProvider.SignResult: # pragma: no cover
        return self.sign_ec(op)

    @_sign.register
    def _sign_eddsa(self, op: EdDSAOperation) -> IProvider.SignResult: # pragma: no cover
        return self.sign_eddsa(op)

    @_sign.register
    def _sign_hmac(self, op: HMACOperation) -> IProvider.SignResult: # pragma: no cover
        return self.sign_hmac(op)

    @_sign.register
    def _sign_rsa(self, op: RSAOperation) -> IProvider.SignResult: # pragma: no cover
        return self.sign_rsa(op)

    def sign_ec(self, op: EllipticCurveOperation) -> IProvider.SignResult:
        """Create a digital signature using an elliptic curve private key."""
        raise NotImplementedError(
            f"Elliptic curve signing is not supported by {type(self).__name__}"
        )

    def sign_eddsa(self, op: EdDSAOperation) -> IProvider.SignResult:
        """Create a digital signature using an Ed448 or Ed25519 private key."""
        raise NotImplementedError(
            f"EdDSA signing is not supported by {type(self).__name__}"
        )

    def sign_hmac(self, op: HMACOperation) -> IProvider.SignResult:
        """Create a digital signature using a HMAC."""
        raise NotImplementedError(
            f"HMAC signing is not supported by {type(self).__name__}"
        )

    def sign_rsa(self, op: RSAOperation) -> IProvider.SignResult:
        """Create a digital signature using an RSA private key."""
        raise NotImplementedError(
            f"RSA signing is not supported by {type(self).__name__}"
        )

    def verify(
        self,
        key: KeySpecification,
        algorithm: str,
        message: types.Digest | types.Message,
        signature: bytes
    ) -> bool | Awaitable[bool]:
        """Verify a digital signature, specified by the operation."""
        op = self.algorithm_operations[algorithm].fromalgorithm(
            algorithm=algorithm,
            spec=key,
            content=message,
            signature=signature
        )
        return self._verify(op)

    @functools.singledispatchmethod # type: ignore
    def _verify(
        self,
        op: SigningOperation
    ) -> bool | Awaitable[bool]:
        raise NotImplementedError(
            f"Operation {type(op).__name__} is not implemented by "
            f"{type(self).__name__}."
        )

    def verify_ec(self, op: types.EllipticCurveOperation) -> bool | Awaitable[bool]:
        """Verify a digital signature that was created with an elliptic curve
        private key.
        """
        raise NotImplementedError(
            f"Elliptic curve verification is not supported by {type(self).__name__}"
        )

    def verify_eddsa(self, op: types.EdDSAOperation) -> bool | Awaitable[bool]:
        """Verify a digital signature that was created with an Ed448 or Ed25519
        private key.
        """
        raise NotImplementedError(
            f"EdDSA verification is not supported by {type(self).__name__}"
        )

    def verify_hmac(self, op: types.HMACOperation) -> bool | Awaitable[bool]:
        """Verify a digital signature that was created with a Hash-based
        Message Authentication Code (HMAC).
        """
        raise NotImplementedError(
            f"HMAC verification is not supported by {type(self).__name__}"
        )

    def verify_rsa(self, op: types.RSAOperation) -> bool | Awaitable[bool]:
        """Verify a digital signature that was created with a Hash-based
        Message Authentication Code (HMAC).
        """
        raise NotImplementedError(
            f"RSA verification is not supported by {type(self).__name__}"
        )

    @_verify.register
    def _verify_ec(
        self,
        op: types.EllipticCurveOperation
    ) -> bool | Awaitable[bool]:
        return self.verify_ec(op)

    @_verify.register
    def _verify_eddsa(
        self,
        op: types.EdDSAOperation
    ) -> bool | Awaitable[bool]:
        return self.verify_eddsa(op)

    @_verify.register
    def _verify_hmac(
        self,
        op: types.HMACOperation
    ) -> bool | Awaitable[bool]:
        return self.verify_hmac(op)

    @_verify.register
    def _verify_rsa(
        self,
        op: types.RSAOperation
    ) -> bool | Awaitable[bool]:
        return self.verify_rsa(op)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`CipherText`."""
from typing import Any

from .data import Data


class CipherText(Data):
    """The result of an encryption operation."""
    __module__: str = 'ckms.types'
    aad: bytes | None
    content: bytes
    iv: bytes | None
    tag: bytes | None

    def __init__(
        self,
        *,
        aad: bytes | None = None,
        iv: bytes | None = None,
        tag: bytes | None = None,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.aad = aad
        self.iv = iv
        self.tag = tag



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic
from pyasn1_modules import rfc2315

from ckms.ext.rfc2459.types import RelativeDistinguishedName


class IssuerAndSerialNumber(pydantic.BaseModel):
    name: list[RelativeDistinguishedName] = pydantic.Field(..., alias='issuer')
    serial_number: int = pydantic.Field(..., alias='serialNumber')

    @pydantic.validator('name', pre=True)
    def validate_name(
        cls,
        value: rfc2315.Name | RelativeDistinguishedName
    ) -> list[RelativeDistinguishedName]:
        names: list[RelativeDistinguishedName] = []
        if isinstance(value, rfc2315.Name):
            for rdn in value[0]: # type: ignore
                names.append(
                    RelativeDistinguishedName.from_attrs(rdn) # type: ignore
                )
            value = names # type: ignore
        assert isinstance(value, list)
        return value


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from .base import Algorithm
from .nullalgorithm import NullAlgorithm


__all__: list[str] = [
    'register',
    'Algorithm',
    'NullAlgorithm'
]


def alias(name: str, alias_of: str) -> None:
    """Create an alias of a cryptographic algorithm."""
    return Algorithm.alias(name, alias_of)


def register(name: str, base: type[Algorithm], params: Any) -> None:
    """Register a new algorithm preset using the given class."""
    return Algorithm.register(name, base=base, **params)


register('null', NullAlgorithm, {})
register('EdDSA', Algorithm, {
    'kty': 'OKP',
    'use': 'sig',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256K', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256K',
    'digest': 'sha256',
    'digest_oid': '1.2.840.10045.4.3.2',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256K-384', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256K',
    'digest': 'sha384',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256K-512', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256K',
    'digest': 'sha512',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256',
    'digest': 'sha256',
    'digest_oid': '1.2.840.10045.4.3.2',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256-384', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256',
    'digest': 'sha384',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES256-512', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-256',
    'digest': 'sha512',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES384-256', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-384',
    'digest': 'sha256',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES384', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-384',
    'digest': 'sha384',
    'digest_oid': '1.2.840.10045.4.3.3',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES384-512', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-384',
    'digest': 'sha512',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES512-256', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-521',
    'digest': 'sha256',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES512-384', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-521',
    'digest': 'sha384',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('ES512', Algorithm, {
    'kty': 'EC',
    'use': 'sig',
    'crv': 'P-521',
    'digest': 'sha512',
    'digest_oid': '1.2.840.10045.4.3.4',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('HS256', Algorithm, {
    'kty': 'oct',
    'use': 'sig',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('HS384', Algorithm, {
    'kty': 'oct',
    'use': 'sig',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'},
    'default': True
})

register('HS512', Algorithm, {
    'kty': 'oct',
    'use': 'sig',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('PS256', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PSS',
    'digest': 'sha256',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('PS384', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PSS',
    'digest': 'sha384',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'},
    'default': True
})

register('PS512', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PSS',
    'digest': 'sha512',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('RS256', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PKCSv1',
    'digest': 'sha256',
    'digest_oid': '1.2.840.113549.1.1.11',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('RS384', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PKCSv1',
    'digest': 'sha384',
    'digest_oid': '1.2.840.113549.1.1.12',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('RS512', Algorithm, {
    'kty': 'RSA',
    'use': 'sig',
    'padding': 'PKCSv1',
    'digest': 'sha512',
    'digest_oid': '1.2.840.113549.1.1.13',
    'allowed_ops': {'sign'},
    'default_ops': {'sign'}
})

register('A128GCM', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'decrypt', 'encrypt'},
    'default_ops': {'decrypt', 'encrypt'}
})

register('A192GCM', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'decrypt', 'encrypt'},
    'default_ops': {'decrypt', 'encrypt'}
})

register('A256GCM', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'decrypt', 'encrypt'},
    'default_ops': {'decrypt', 'encrypt'}
})

register('A128KW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+RFC3394',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'}
})

register('A192KW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+RFC3394',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'}
})

register('A256KW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+RFC3394',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'}
})

register('A128GCMKW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'}
})

register('A192GCMKW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'}
})

register('A256GCMKW', Algorithm, {
    'kty': 'oct',
    'use': 'enc',
    'mode': 'AES+GCM',
    'allowed_ops': {'unwrapKey', 'wrapKey'},
    'default_ops': {'unwrapKey', 'wrapKey'},
    'default': True
})

register('ECDH-ES', Algorithm, {
    'kty': 'OKP',
    'use': 'enc',
    'allowed_ops': {'deriveKey'},
    'default_ops': {'deriveKey'}
})

register('ECDH-ES+A128KW', Algorithm, {
    'kty': 'OKP',
    'use': 'enc',
    'allowed_ops': {'deriveKey'},
    'default_ops': {'deriveKey'}
})

register('ECDH-ES+A192KW', Algorithm, {
    'kty': 'OKP',
    'use': 'enc',
    'allowed_ops': {'deriveKey'},
    'default_ops': {'deriveKey'}
})

register('ECDH-ES+A256KW', Algorithm, {
    'kty': 'OKP',
    'use': 'enc',
    'allowed_ops': {'deriveKey'},
    'default_ops': {'deriveKey'},
    'default': True
})

register('RSA1_5', Algorithm, {
    'kty': 'RSA',
    'use': 'enc',
    'padding': 'PKCSv1',
    'digest': 'sha1',
    'allowed_ops': {'unwrapKey', 'decrypt'},
    'default_ops': {'unwrapKey'}
})

register('RSA-OAEP', Algorithm, {
    'kty': 'RSA',
    'use': 'enc',
    'padding': 'OAEP',
    'digest': 'sha1',
    'allowed_ops': {'unwrapKey', 'decrypt'},
    'default_ops': {'unwrapKey'}
})

register('RSA-OAEP-256', Algorithm, {
    'kty': 'RSA',
    'use': 'enc',
    'padding': 'OAEP',
    'digest': 'sha256',
    'allowed_ops': {'unwrapKey', 'decrypt'},
    'default_ops': {'unwrapKey'}
})

register('RSA-OAEP-384', Algorithm, {
    'kty': 'RSA',
    'use': 'enc',
    'padding': 'OAEP',
    'digest': 'sha384',
    'allowed_ops': {'unwrapKey', 'decrypt'},
    'default_ops': {'unwrapKey'},
    'default': True
})

register('RSA-OAEP-512', Algorithm, {
    'kty': 'RSA',
    'use': 'enc',
    'padding': 'OAEP',
    'digest': 'sha512',
    'allowed_ops': {'unwrapKey', 'decrypt'},
    'default_ops': {'unwrapKey'}
})


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic

from ckms.ext.rfc5758.types import SignatureDigestAlgorithmType
from .algorithmidentifier import AlgorithmIdentifier
from .subjectpublickeyinfo import SubjectPublicKeyInfo


class SignatureDigestAlgorithmIdentifier(AlgorithmIdentifier):
    name: SignatureDigestAlgorithmType = pydantic.Field(..., alias='algorithm') # type: ignore

    def get_qualname(self, key: SubjectPublicKeyInfo) -> str:
        """Return the qualified name as used internally with :mod:`ckms`."""
        return self.name.qualname()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import ckms.core
from ckms.core.models import KeySpecification


ENCRYPTION_KEYS: list[KeySpecification] = []

SIGNING_KEYS_ASYMMETRIC: list[KeySpecification] = [
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'vault': 'unimatrixdev',
        'name': 'ec-sign-p256-sha256'
    }),
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'vault_url': 'https://unimatrixdev.vault.azure.net/',
        'name': 'ec-sign-p384-sha384'
    }),
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'vault_url': 'https://unimatrixdev.vault.azure.net/',
        'name': 'ec-sign-p521-sha512'
    }),
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'algorithm': 'RS256',
        'vault': 'unimatrixdev',
        'name': 'rsa-sign-pkcs1-2048-sha256'
    }),
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'algorithm': 'RS384',
        'vault': 'unimatrixdev',
        'name': 'rsa-sign-pkcs1-2048-sha384'
    }),
    ckms.core.parse_spec({
        'provider': 'azure',
        'use': 'sig',
        'algorithm': 'RS512',
        'vault': 'unimatrixdev',
        'name': 'rsa-sign-pkcs1-2048-sha512'
    }),
]

SIGNING_KEYS_SYMMETRIC: list[KeySpecification] = []

SIGNING_KEYS: list[KeySpecification] = [
    *SIGNING_KEYS_ASYMMETRIC
]

SUPPORTED_KEYS: list[KeySpecification] = [
    *SIGNING_KEYS
]

@pytest.fixture(scope='session')
def supported_keys() -> list[KeySpecification]:
    return SUPPORTED_KEYS


# pylint: skip-file
# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .basekeyspecification import BaseKeySpecification
from .contentencryptionkey import ContentEncryptionKey
from .joseheader_ import JOSEHeader
from .joseobject import JOSEObject
from .jsonwebencryption import JSONWebEncryption
from .jsonwebencryptionflattened import JSONWebEncryptionFlattened
from .jsonwebencryptionwithrecipients import JSONWebEncryptionWithRecipients
from .jsonwebsignature import JSONWebSignature
from .keyspecification import KeySpecification
from .publickeyspecification import PublicKeySpecification
from .remoteblobparams import RemoteBlobParams
from .signature import Signature


__all__ = [
    'BaseKeySpecification',
    'ContentEncryptionKey',
    'JOSEHeader',
    'JOSEObject',
    'JSONWebEncryption',
    'JSONWebEncryptionFlattened',
    'JSONWebEncryptionWithRecipients',
    'JSONWebSignature',
    'KeySpecification',
    'PublicKeySpecification',
    'RemoteBlobParams',
    'Signature'
]



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONWebSignature`."""
import json
from typing import Any
from typing import Generator

from ckms.types import CipherText
from ckms.types import ClaimSet
from ckms.types import Data
from ckms.core.const import DEFAULT_CONTENT_ENCRYPTION_ALGORITHM
from ckms.core.models import ContentEncryptionKey
from ckms.utils import b64encode
from ckms.utils import b64encode_str
from ckms.utils import b64encode_json
from ckms.types import Encrypter
from ckms.types import Signer
from ckms.types.icontentencryptionkey import IContentEncryptionKey

from .recipient import Recipient
from .signature_ import Signature


class JOSEPayload:
    __module__: str = 'ckms.jose'
    _allow_compact: bool
    _allow_flattened: bool
    _aad: bytes | None = None
    _ciphertext: CipherText | None = None
    _claims: ClaimSet | None = None
    _content: bytes | None
    _content_type: str
    _direct: bool = False
    _encryption: str
    _encryption_key: ContentEncryptionKey | Encrypter
    _payload: Data
    _protected: dict[str, Any]
    _recipients: list[Recipient]
    _signatures: list[Signature]

    @property
    def aad(self) -> bytes:
        return (
            b64encode_json(self._protected)
            if not self._aad
            else b'.'.join([
                b64encode_json(self._protected),
                b64encode(self._aad)
            ])
        )

    def __init__(
        self,
        payload: dict[str, Any] | ClaimSet | Data,
        content_type: str | None = None,
        encryption: str = DEFAULT_CONTENT_ENCRYPTION_ALGORITHM,
        encryption_key: Encrypter | None = None,
        protected: dict[str, Any] | None = None,
        aad: bytes | dict[str, Any] |  None = None,
        allow_compact: bool = True,
        allow_flattened: bool = True
    ):
        """Initialize a new :class:`JOSEObject` instance.
        
        Args:
            payload: a :class:`ckms.types.ClaimSet` instance holding claims.
                This implies that the payload is a JSON Web Token (JWT).
            content_type (bytes): indicates the content type of the payload.
                If the payload is a JWT, then this defaults to ``'JWT'``.
        """
        if isinstance(payload, dict):
            payload = ClaimSet.parse_obj(payload)
        if isinstance(payload, ClaimSet):
            self._claims = payload
            payload = Data(
                buf=str.encode(payload.json(exclude_defaults=True), 'utf-8'),
                content_type=content_type or "JWT"
            )
        if isinstance(aad, dict):
            aad = b64encode_json(aad)
        self._aad = aad
        self._allow_compact = allow_compact
        self._allow_flattened = allow_flattened
        self._content = None
        self._encryption = encryption
        self._payload = payload
        self._protected = protected or {}
        self._recipients = []
        self._signatures = []
        if payload.content_type is None:
            raise ValueError("Data.content_type must not be None.")
        self._content_type = payload.content_type
        self._encryption_key = encryption_key or ContentEncryptionKey.generate(encryption)

    def add_recipient(
        self,
        *,
        encrypter: Encrypter,
        header: dict[str, Any] | None = None,
        direct: bool = False
    ) -> Recipient:
        """Encrypt the Content Encryption Key (CEK) for the given recipient.

        If Direct Encryption or Direct Key Agreement is used, set `encrypter`
        as the content encryption key if there are no other recipients.
        Otherwise, raise an exception, as there can be no multiple recipients
        when using these modes.

        The `header` parameter specifies a dictionary holding the per-recipient
        header. The set of keys in this header must be disjoint from the keys
        that are added to other headers, being at least:

        - ``typ``
        - ``enc``        
        """
        assert encrypter.algorithm is not None
        header = header or {}
        if encrypter.algorithm in ('ECDH-ES',):
            raise NotImplementedError
        if direct:
            if len(self._recipients) >= 1:
                raise ValueError(
                    "Can not use Direct Encryption or Direct Key Agreement with "
                    "more than one recipient."
                )
            if not encrypter.can_encrypt():
                raise ValueError("Key not suitable for encryption.")
            # Ensure that the _encryption matches the algorithm of the
            # key.
            self._encryption = encrypter.algorithm
            self._encryption_key = encrypter
        self._recipients.append(
            Recipient(encrypter=encrypter, header=header)
        )
        return self._recipients[-1]

    def add_signature(
        self,
        *,
        signer: Signer,
        protected: dict[str, Any] | None = None,
        header: dict[str, Any] | None = None
    ) -> Signature:
        """Sign the :class:`JOSEPayload` with the given signer."""
        if self.is_encrypted():
            raise ValueError("Signing must occur prior to encryption.")
        protected = protected or {}
        protected.setdefault('alg', signer.algorithm)
        protected.setdefault('kid', signer.kid)
        if self.is_jwt():
            protected['typ'] = self._content_type
        else:
            protected.setdefault('cty', self._content_type)
        sig = Signature(
            protected=protected,
            signer=signer,
            header=header,
            payload=self._payload
        )
        self._signatures.append(sig)
        return sig

    def can_flatten(self) -> bool:
        """Return a boolean indicating if the top-level object can be
        flattened.
        """
        return (
            not self.is_multirecipient()
            if self.is_encrypted()
            else not self.is_multisigned()
        )

    def get_payload(self) -> bytes:
        """Return the payload of a JWE or JWS."""
        # If JOSEPayload._content is None here, then there are no signers and
        # no recipients, as the add_signature() or add_recipient() methods
        # were never called.
        return self._content or b64encode(bytes(self._payload))

    def is_direct(self) -> bool:
        """Return a boolean indicating if the JWE uses either
        Direct Encryption or Direct Key Agreement.
        """
        return not isinstance(self._encryption_key, IContentEncryptionKey)

    def is_finalized(self) -> bool:
        """Return a boolean indicating if all signatories and recipients
        have executed their cryptographic operations.
        """
        return all(self._recipients) and all(map(bool, self._signatures))

    def is_nested(self) -> bool:
        """Return a boolean indicating if this is a nested object e.g.
        JWS inside JWE.
        """
        return bool(self._signatures) and bool(self._recipients)

    def is_encrypted(self) -> bool:
        """Return a boolean indicating if this object is encrypted."""
        return bool(self._recipients)

    def is_jwt(self) -> bool:
        """Return a boolean indicating if the payload is a JSON Web Token (JWT)."""
        return self._claims is not None

    def is_multirecipient(self) -> bool:
        """Return a boolean indicating if the object has multiple recipients."""
        return len(self._recipients) > 1

    def is_multisigned(self) -> bool:
        """Return a boolean indicating if the object has multiple signatures."""
        return len(self._signatures) > 1

    def is_signed(self) -> bool:
        """Return a boolean indicating if this object is signed."""
        return bool(self._signatures)

    async def serialize(
        self,
        compact: bool = False,
        flatten: bool = True,
        aad: bytes | dict[str, Any] | None = None,
        compress: bool = False,
        protected: dict[str, Any] | None = None
    ) -> str:
        """Serializes the JOSE object. Determine if there are recipients
        (object is JWE), signatures (object is JWS) or both (object is nested).
        Return a string containing the JOSE object, which will be JWS if there
        are signatures and JWE if there are recipients. If there are signatures
        and recipients, the string containing a nested object (signed then
        encrypted).

        A serialization format may be specified with the `compact` and `flattened`
        parameters. These are only availabe if there is a single signature and/or
        recipient. If `compact` is True, then JWE/JWS Compact Serialization is
        used. Otherwise, JWE/JWS Flattened Serialization is used, unless `flatten`
        is ``False``.

        If the encryption algorithm supports Authenticated Encryption with
        Associated Data (AEAD), the `aad` parameter may be used to specify
        Additional Authenticated Data (AAD).

        Additional protected header claims can be supplied with the `protected`
        parameter. The claims are added to the header of the top-level JOSE
        object (i.e. if the object is nested, the inner JWS does not receive
        the protected claims).

        The compress paramater is a boolean indicating that the payload of a
        JWE must be compressed using the DEFLATE (:rfc:`1951`) algorithm. If
        the instance is not JWE (i.e. does not have recipients), this
        parameter is ignored.
        """
        if compact and (self.is_multirecipient() or self.is_multisigned()):
            raise ValueError(
                "JWE/JWS Compact Serialization is not available when there "
                "are multiple signatures or recipients."
            )

        if compact and (aad or self._aad):
            raise ValueError(
                "Additional Authentication Data (AAD) can not be supplied with "
                "the JWE Compact Serialziation format."
            )

        if compress:
            raise NotImplementedError

        # Create all signatures and encrypt the plain text if there are reci
        obj = await self._payload.as_bytes()
        common_protected = protected or {}
        if self.is_signed():
            obj = b64encode(obj)
            signatures: list[dict[str, str]] = []
            protected = self._get_jws_protected_header(compact=compact)
            if not self.is_encrypted():
                # This object is top-level, add the shared protected header.
                protected.update(common_protected)
            for sig in self._signatures:
                await sig.sign(obj, protected=protected)
                signatures.append(sig.dict(compact=compact))
            obj = self._serialize_jws(
                obj={
                    'payload': bytes.decode(obj, 'ascii'),
                    'signatures': signatures
                },
                compact=compact,
                flatten=flatten
            )
        if self.is_encrypted():
            if isinstance(obj, str):
                obj = str.encode(obj, 'utf-8')
            recipients: list[dict[str, Any]] = []
            for recipient in self._recipients:
                await recipient.encrypt(cek=self._encryption_key) # type: ignore
                recipients.append(recipient.dict(compact=compact, flatten=flatten))

            aad = aad or self._aad
            if aad and isinstance(aad, dict):
                aad = str.encode(json.dumps(aad), 'utf-8')

            # With compact serialization, there is a single recipient
            # and its header values need to be added to the protected
            # header, since the unprotected and per-recipient headers
            # are not included in the serialization.
            if compact:
                common_protected.update(recipients[0].get('header') or {})

            protected_b64 = b64encode_json({
                **self._get_jwe_protected_header(compact=compact),
                **common_protected
            })

            assert aad is None or isinstance(aad, bytes)
            ct = await self._encryption_key.encrypt(
                pt=obj,
                aad=self._get_aad(protected_b64, aad)
            )
            obj = self._serialize_jwe(
                obj={
                    'protected': protected_b64.decode('ascii'),
                    'unprotected': {},
                    'iv': b64encode_str(ct.iv or b''),
                    'ciphertext': b64encode_str(bytes(ct)),
                    'tag': b64encode_str(ct.tag or b''),
                    'aad': b64encode_str(aad or b''),
                    'recipients': recipients
                },
                compact=compact,
                flatten=flatten
            )

        assert obj is not None
        assert isinstance(obj, str), obj
        return obj

    def _get_jwe_protected_header(self, compact: bool = False) -> dict[str, Any]:
        # The "typ" value "JOSE" can be used by applications to indicate that
        # this object is a JWS or JWE using the JWS Compact Serialization or
        # the JWE Compact Serialization.  The "typ" value "JOSE+JSON" can be
        # used by applications to indicate that this object is a JWS or JWE
        # using the JWS JSON Serialization or the JWE JSON Serialization.
        # Other type values can also be used by applications (RFC 7515, 4.1.9).
        protected = self._protected = {
            **self._protected,
            'cty': self._content_type,
            'enc': self._encryption,
            'typ': "JOSE+JSON"
        }
        if compact:
            protected['typ'] = "JOSE"

        if self.is_signed():
            protected['cty'] = "JOSE" if compact else "JOSE+JSON"

        if self.is_jwt():
            # JSON Web Tokens (JWTs) have some special rules regarding the "typ"
            # and "cty" header values.
            del protected['cty'] # cty header is only included when nesting.
            protected['typ'] = self._content_type
            if self.is_nested():
                # In the case that nested signing or encryption is employed,
                # this Header Parameter MUST be present; in this case, the
                # value MUST be "JWT", to indicate that a Nested JWT is carried
                # in this JWT.  While media type names are not case sensitive,
                # it is RECOMMENDED that "JWT" always be spelled using
                # uppercase characters for compatibility with legacy
                # implementations (RFC 7519, 5.2).
                #
                # The actual content type specified in the innner JWS.
                protected['typ'] = "JWT"
                protected['cty'] = "JWT"

        return {k: v for k, v in protected.items() if v}

    def _get_aad(self, protected: bytes, aad: bytes | None) -> bytes:
        return (
            protected
            if aad is None
            else b'.'.join([
                protected,
                b64encode(aad)
            ])
        )

    def _serialize_jwe(
        self,
        obj: dict[str, Any],
        compact: bool,
        flatten: bool
    ) -> str:
        compact = compact and self._allow_compact
        flatten = flatten and self._allow_flattened
        jwe: dict[str, Any] | str = obj
        recipients = jwe.get('recipients') or []
        if len(recipients) == 1 and (flatten or compact):
            jwe.update(jwe['recipients'][0])
            if compact:
                # This implementation always sets the protected header, holding the
                # enc and typ claims.
                jwe = (
                    f"{jwe['protected']}."
                    f"{jwe['encrypted_key']}."
                    f"{jwe.get('iv') or ''}."
                    f"{jwe['ciphertext']}."
                    f"{jwe.get('tag') or ''}"
                )
        if isinstance(jwe, dict):
            jwe = json.dumps({k: v for k, v in jwe.items() if v}, sort_keys=True)
        return jwe

    def _get_jws_protected_header(self, compact: bool = False) -> dict[str, Any]:
        protected = {'typ': "JOSE+JSON"}
        if compact:
            protected['typ'] = "JOSE"
        if self.is_jwt():
            protected['typ'] = self._content_type or "JWT"
        return protected

    def _serialize_jws(
        self,
        obj: dict[str, Any],
        compact: bool,
        flatten: bool
    ) -> str:
        compact = compact and self._allow_compact
        flatten = flatten and self._allow_flattened
        jws: dict[str, Any] | str = obj
        signatures = jws.get('signatures') or []
        if len(signatures) == 1 and (flatten or compact):
            jws.update(signatures[0])
            if 'header' in obj and not obj.get('header'):
                del jws['header']
            if compact:
                jws = f"{jws['protected']}.{jws['payload']}.{jws['signature']}"
        if not compact:
            jws = json.dumps(obj, sort_keys=True)
        assert isinstance(jws, str)
        return jws

    def __await__(self) -> Generator[Any, None, str]:
        compact = not self.is_multirecipient() and not self.is_multisigned()
        return self.serialize(compact=compact).__await__()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import Optional

from pyasn1.type.base import Asn1Type
from pyasn1.type import univ
from pyasn1_modules import rfc5280

from .relativedistinguishedname import RelativeDistinguishedName


class RDNSequence(list[RelativeDistinguishedName]):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: list[RelativeDistinguishedName] | univ.SetOf | None
    ) -> Optional['RDNSequence']:
        if isinstance(value, rfc5280.Name):
            value = [
                RelativeDistinguishedName.validate(x)
                for x in value['rdnSequence']
            ]
        elif isinstance(value, Asn1Type):
            raise NotImplementedError(type(value))
        return cls(value) if value is not None else None

    def asn1(self) -> rfc5280.Name:
        obj = rfc5280.Name()
        seq = rfc5280.RDNSequence()
        for rdn in self:
            seq.append(rdn.asn1())
        obj['rdnSequence'] = seq
        return obj


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime

import pydantic

from pyasn1.type.useful import GeneralizedTime
from pyasn1.type import univ
from pyasn1_modules import rfc2315
from pyasn1_modules import rfc3161

from ckms.ext import asn1
from ckms.ext.rfc3161.oid import TSTInfoOID
from .messageimprint import MessageImprint


class TSTInfo(pydantic.BaseModel):
    version: int
    policy: str | None
    serial_number: int = pydantic.Field(..., alias='serialNumber')
    generated: datetime.datetime = pydantic.Field(..., alias='genTime')
    ordering: bool | None
    nonce: int | None
    message_imprint: MessageImprint = pydantic.Field(..., alias='messageImprint')
    #tsa: str | None

    @classmethod
    def from_signed_data(cls, content_info: rfc2315.ContentInfo) -> 'TSTInfo':
        oid = str(content_info['contentType']) # type: ignore
        if str(oid) != TSTInfoOID:
            raise ValueError(f'Unexpected content type: {oid}')
        return cls.parse_obj(
            asn1.decode_ber(
                value=content_info['content'], # type: ignore
                spec=rfc3161.TSTInfo()
            )
        )

    @pydantic.validator('generated', pre=True)
    def validate_generated(cls, value: datetime.datetime | GeneralizedTime) -> datetime.datetime | str:
        if isinstance(value, GeneralizedTime):
            value = datetime.datetime.strptime(str(value), '%Y%m%d%H%M%SZ')
        return value

    @pydantic.validator('nonce', pre=True)
    def validate_nonce(cls, value: int | univ.Integer | None) -> int | None:
        if isinstance(value, univ.Integer):
            value = int(value) if value.isValue else None
        return value

    @pydantic.validator('policy', pre=True)
    def validate_policy(cls, value: str | rfc3161.TSAPolicyId | None) -> str | None:
        if isinstance(value, rfc3161.TSAPolicyId):
            value = str(value) if value.isValue else None
        return value

    @pydantic.validator('serial_number', pre=True)
    def validate_serial_number(cls, value: int | univ.Integer) -> int:
        if isinstance(value, univ.Integer):
            value = int(value)
        return value

    #@pydantic.validator('tsa', pre=True)
    #def validate_tsa(cls, value: str | rfc3161.GeneralName) -> str:
    #    if isinstance(value, rfc3161.GeneralName):
    #        value = str(value)
    #    return value